# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Executor for running inside a container.

Useful for Kubernetes and TES batch systems.
"""
import base64
import logging
import os
import pickle
import subprocess
import sys
from typing import Any, Dict, List, Optional

from toil.batchSystems.abstractBatchSystem import EXIT_STATUS_UNAVAILABLE_VALUE
from toil.job import JobDescription
from toil.resource import Resource
from toil.statsAndLogging import configure_root_logger, set_log_level

logger = logging.getLogger(__name__)


def pack_job(job_desc: JobDescription, user_script: Optional[Resource] = None, environment: Optional[Dict[str, str]] = None) -> List[str]:
    """
    Create a command that, when run, will execute the given job.

    :param job_desc: Job description for the job to run.
    :param user_script: User script that will be loaded before the job is run.
    :param environment: Environment variable dict that will be applied before
    the job is run.

    :returns: Command to run the job, as an argument list that can be run
    inside the Toil appliance container.
    """
    # Make a job dict to send to the executor.
    # TODO: Factor out executor setup from here and Kubernetes and TES
    job: Dict[str, Any] = {"command": job_desc.command}
    if user_script is not None:
        # If there's a user script resource be sure to send it along
        job['userScript'] = user_script
    if environment is not None:
        # We also may have an environment to send.
        job['environment'] = environment
    # Encode it in a form we can send in a command-line argument. Pickle in
    # the highest protocol to prevent mixed-Python-version workflows from
    # trying to work. Make sure it is text so we can ship it via JSON.
    encoded_job = base64.b64encode(pickle.dumps(job, pickle.HIGHEST_PROTOCOL)).decode('utf-8')
    # Make a command to run it in the executor
    command_list = ['_toil_contained_executor', encoded_job]

    return command_list


def executor() -> None:
    """
    Main function of the _toil_contained_executor entrypoint.

    Runs inside the Toil container.

    Responsible for setting up the user script and running the command for the
    job (which may in turn invoke the Toil worker entrypoint).

    """

    configure_root_logger()
    set_log_level("DEBUG")
    logger.debug("Starting executor")

    # If we don't manage to run the child, what should our exit code be?
    exit_code = EXIT_STATUS_UNAVAILABLE_VALUE

    if len(sys.argv) != 2:
        logger.error('Executor requires exactly one base64-encoded argument')
        sys.exit(exit_code)

    # Take in a base64-encoded pickled dict as our first argument and decode it
    try:
        # Make sure to encode the text arguments to bytes before base 64 decoding
        job = pickle.loads(base64.b64decode(sys.argv[1].encode('utf-8')))
    except:
        exc_info = sys.exc_info()
        logger.error('Exception while unpickling task: ', exc_info=exc_info)
        sys.exit(exit_code)

    if 'environment' in job:
        # Adopt the job environment into the executor.
        # This lets us use things like TOIL_WORKDIR when figuring out how to talk to other executors.
        logger.debug('Adopting environment: %s', str(job['environment'].keys()))
        for var, value in job['environment'].items():
            os.environ[var] = value

    # Set JTRES_ROOT and other global state needed for resource
    # downloading/deployment to work.
    # TODO: Every worker downloads resources independently.
    # We should have a way to share a resource directory.
    logger.debug('Preparing system for resource download')
    Resource.prepareSystem()
    try:
        if 'userScript' in job:
            job['userScript'].register()

        # Start the child process
        logger.debug("Invoking command: '%s'", job['command'])
        child = subprocess.Popen(job['command'],
                                 preexec_fn=lambda: os.setpgrp(),
                                 shell=True)

        # Reproduce child's exit code
        exit_code = child.wait()
    except:
        # This will print a traceback for us, since exit() in the finally
        # will bypass the normal way of getting one.
        logger.exception('Encountered exception running child')
    finally:
        logger.debug('Cleaning up resources')
        # TODO: Change resource system to use a shared resource directory for everyone.
        # Then move this into worker cleanup somehow
        Resource.cleanSystem()
        logger.debug('Shutting down')
        sys.exit(exit_code)





# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Delete a job store used by a previous Toil workflow invocation."""
import logging

from toil.common import Toil, parser_with_common_options
from toil.jobStores.abstractJobStore import NoSuchJobStoreException
from toil.statsAndLogging import set_logging_from_options

logger = logging.getLogger(__name__)


def main() -> None:
    parser = parser_with_common_options(jobstore_option=True)

    options = parser.parse_args()
    set_logging_from_options(options)
    try:
        jobstore = Toil.getJobStore(options.jobStore)
        try:
            # TODO: We need to resume successfully to delete encrypted AWS job
            # stores supposedly. We should avoid needing this.
            # See https://github.com/DataBiosphere/toil/issues/1041
            jobstore.resume()
        except NoSuchJobStoreException:
            logger.warning("Job store does not appear to exist. Deleting anyway.")
        except:
            logger.exception("Job store appears to be in a bad state. Deleting anyway.")
        jobstore.destroy()
        logger.info(f"Successfully deleted the job store: {options.jobStore}")
    except:
        logger.critical(f"Failed to delete the job store: {options.jobStore}")
        raise



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Batch system for running Toil workflows on Kubernetes.

Ony useful with network-based job stores, like AWSJobStore.

Within non-privileged Kubernetes containers, additional Docker containers
cannot yet be launched. That functionality will need to wait for user-mode
Docker
"""
import datetime
import logging
import math
import os
from queue import Empty, Queue
import string
import sys
import tempfile
from threading import Event, Thread, Condition, RLock
import time
import uuid
from argparse import ArgumentParser, _ArgumentGroup
from typing import (Any,
                    Callable,
                    Dict,
                    Iterator,
                    List,
                    Set,
                    Literal,
                    Optional,
                    Tuple,
                    Type,
                    TypeVar,
                    Union,
                    cast,
                    overload)

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec
if sys.version_info >= (3, 8):
    from typing import Protocol, TypedDict, runtime_checkable
else:
    from typing_extensions import Protocol, TypedDict, runtime_checkable
# TODO: When this gets into the standard library, get it from there and drop
import urllib3
import yaml
# The Right Way to use the Kubernetes module is to `import kubernetes` and then you get all your stuff as like ApiClient. But this doesn't work for the stubs: the stubs seem to only support importing things from the internal modules in `kubernetes` where they are actually defined. See for example <https://github.com/MaterializeInc/kubernetes-stubs/issues/9 and <https://github.com/MaterializeInc/kubernetes-stubs/issues/10>. So we just import all the things we use into our global namespace here.
from kubernetes.client import (BatchV1Api,
                               CoreV1Api,
                               CustomObjectsApi,
                               V1Affinity,
                               V1Container,
                               V1ContainerStatus,
                               V1EmptyDirVolumeSource,
                               V1HostPathVolumeSource,
                               V1Job,
                               V1JobCondition,
                               V1JobSpec,
                               V1NodeAffinity,
                               V1NodeSelector,
                               V1NodeSelectorRequirement,
                               V1NodeSelectorTerm,
                               V1ObjectMeta,
                               V1Pod,
                               V1PodSpec,
                               V1PodTemplateSpec,
                               V1PreferredSchedulingTerm,
                               V1ResourceRequirements,
                               V1SecretVolumeSource,
                               V1Toleration,
                               V1Volume,
                               V1VolumeMount)
from kubernetes.client.api_client import ApiClient
from kubernetes.client.exceptions import ApiException
from kubernetes.config.config_exception import ConfigException
from kubernetes.config.incluster_config import load_incluster_config
from kubernetes.config.kube_config import (list_kube_config_contexts,
                                           load_kube_config)
# TODO: Watch API is not typed yet
from kubernetes.watch import Watch  # type: ignore
# typing-extensions dependency on Pythons that are new enough.
from typing_extensions import NotRequired

from toil import applianceSelf
from toil.batchSystems.abstractBatchSystem import (EXIT_STATUS_UNAVAILABLE_VALUE,
                                                   BatchJobExitReason,
                                                   InsufficientSystemResources,
                                                   ResourcePool,
                                                   UpdatedBatchJobInfo)
from toil.batchSystems.cleanup_support import BatchSystemCleanupSupport
from toil.batchSystems.contained_executor import pack_job
from toil.batchSystems.options import OptionSetter
from toil.common import Config, Toil, SYS_MAX_SIZE
from toil.job import JobDescription, Requirer
from toil.lib.conversions import human2bytes
from toil.lib.misc import get_user_name, slow_down, utc_now
from toil.lib.retry import ErrorCondition, retry
from toil.resource import Resource

logger = logging.getLogger(__name__)
retryable_kubernetes_errors: List[Union[Type[Exception], ErrorCondition]] = [
    urllib3.exceptions.MaxRetryError,
    urllib3.exceptions.ProtocolError,
    ApiException
]


def is_retryable_kubernetes_error(e: Exception) -> bool:
    """
    A function that determines whether or not Toil should retry or stop given
    exceptions thrown by Kubernetes.
    """
    for error in retryable_kubernetes_errors:
        if isinstance(error, type) and isinstance(e, error):
            return True
    return False

# Represents a collection of label or taint keys and their sets of acceptable (or unacceptable) values.
KeyValuesList = List[Tuple[str, List[str]]]

class KubernetesBatchSystem(BatchSystemCleanupSupport):
    @classmethod
    def supportsAutoDeployment(cls) -> bool:
        return True

    class _ApiStorageDict(TypedDict):
        """
        Type-enforcing dict for our API object cache.
        """

        namespace: NotRequired[str]
        batch: NotRequired[BatchV1Api]
        core: NotRequired[CoreV1Api]
        customObjects: NotRequired[CustomObjectsApi]


    def __init__(self, config: Config, maxCores: int, maxMemory: int, maxDisk: int) -> None:
        super().__init__(config, maxCores, maxMemory, maxDisk)

        # Re-type the config to make sure it has all the fields we need.
        assert isinstance(config, KubernetesBatchSystem.KubernetesConfig)

        # Turn down log level for Kubernetes modules and dependencies.
        # Otherwise if we are at debug log level, we dump every
        # request/response to Kubernetes, including tokens which we shouldn't
        # reveal on CI.
        logging.getLogger('kubernetes').setLevel(logging.ERROR)
        logging.getLogger('requests_oauthlib').setLevel(logging.ERROR)

        # This will hold the last time our Kubernetes credentials were refreshed
        self.credential_time: Optional[datetime.datetime] = None
        # And this will hold our cache of API objects
        self._apis: KubernetesBatchSystem._ApiStorageDict = {}

        # Get our namespace (and our Kubernetes credentials to make sure they exist)
        self.namespace = self._api('namespace')

        # Decide if we are going to mount a Kubernetes host path as the Toil
        # work dir in the workers, for shared caching.
        self.host_path = config.kubernetes_host_path

        # Get the service account name to use, if any.
        self.service_account = config.kubernetes_service_account

        # Get how long we should wait for a pod that lands on a node to
        # actually start.
        self.pod_timeout = config.kubernetes_pod_timeout

        # Get the username to mark jobs with
        username = config.kubernetes_owner
        # And a unique ID for the run
        self.unique_id = uuid.uuid4()

        # Create a prefix for jobs, starting with our username
        self.job_prefix = f'{username}-toil-{self.unique_id}-'
        # Instead of letting Kubernetes assign unique job names, we assign our
        # own based on a numerical job ID. This functionality is managed by the
        # BatchSystemLocalSupport.

        # The UCSC GI Kubernetes cluster (and maybe other clusters?) appears to
        # relatively promptly destroy finished jobs for you if they don't
        # specify a ttlSecondsAfterFinished. This leads to "lost" Toil jobs,
        # and (if your workflow isn't allowed to run longer than the default
        # lost job recovery interval) timeouts of e.g. CWL Kubernetes
        # conformance tests. To work around this, we tag all our jobs with an
        # explicit TTL that is long enough that we're sure we can deal with all
        # the finished jobs before they expire.
        self.finished_job_ttl = 3600  # seconds

        # Here is where we will store the user script resource object if we get one.
        self.user_script: Optional[Resource] = None

        # Ge the image to deploy from Toil's configuration
        self.docker_image = applianceSelf()

        # Try and guess what Toil work dir the workers will use.
        # We need to be able to provision (possibly shared) space there.
        self.worker_work_dir = Toil.getToilWorkDir(config.workDir)
        if (config.workDir is None and
            os.getenv('TOIL_WORKDIR') is None and
            self.worker_work_dir == tempfile.gettempdir()):

            # We defaulted to the system temp directory. But we think the
            # worker Dockerfiles will make them use /var/lib/toil instead.
            # TODO: Keep this in sync with the Dockerfile.
            self.worker_work_dir = '/var/lib/toil'

        # A Toil-managed Kubernetes cluster will have most of its temp space at
        # /var/tmp, which is where really large temp files really belong
        # according to https://systemd.io/TEMPORARY_DIRECTORIES/. So we will
        # set the default temporary directory to there for all our jobs.
        self.environment['TMPDIR'] = '/var/tmp'

        # Get the name of the AWS secret, if any, to mount in containers.
        self.aws_secret_name = os.environ.get("TOIL_AWS_SECRET_NAME", None)

        # Set this to True to enable the experimental wait-for-job-update code
        self.enable_watching = os.environ.get("KUBE_WATCH_ENABLED", False)

        # This will be a label to select all our jobs.
        self.run_id = f'toil-{self.unique_id}'

        # Keep track of available resources.
        maxMillicores = int(SYS_MAX_SIZE if self.maxCores == SYS_MAX_SIZE else self.maxCores * 1000)
        self.resource_sources = [
            # A pool representing available job slots
            ResourcePool(self.config.max_jobs, 'job slots'),
            # A pool representing available CPU in units of millicores (1 CPU
            # unit = 1000 millicores)
            ResourcePool(maxMillicores, 'cores'),
            # A pool representing available memory in bytes
            ResourcePool(self.maxMemory, 'memory'),
            # A pool representing the available space in bytes
            ResourcePool(self.maxDisk, 'disk'),
        ]

        # A set of job IDs that are queued (useful for getIssuedBatchJobIDs())
        self._queued_job_ids: Set[int] = set()

        # Keep track of the acquired resources for each job
        self._acquired_resources: Dict[str, List[int]] = {}

        # Queue for jobs to be submitted to the Kubernetes cluster
        self._jobs_queue: Queue[Tuple[int, JobDescription, V1PodSpec]] = Queue()

        # A set of job IDs that should be killed
        self._killed_queue_jobs: Set[int] = set()

        # We use this event to signal shutdown
        self._shutting_down = Event()

        # A lock to protect critical regions when working with queued jobs.
        self._mutex = RLock()

        # A condition set to true when there is more work to do. e.g.: new job
        # in the queue or any resource becomes available.
        self._work_available = Condition(lock=self._mutex)

        self.schedulingThread = Thread(target=self._scheduler, daemon=True)
        self.schedulingThread.start()

    def _pretty_print(self, kubernetes_object: Any) -> str:
        """
        Pretty-print a Kubernetes API object to a YAML string. Recursively
        drops boring fields.
        Takes any Kubernetes API object; not clear if any base type exists for
        them.
        """

        if not kubernetes_object:
            return 'None'

        # We need a Kubernetes widget that knows how to translate
        # its data structures to nice YAML-able dicts. See:
        # <https://github.com/kubernetes-client/python/issues/1117#issuecomment-939957007>
        api_client: ApiClient = ApiClient()

        # Convert to a dict
        root_dict = api_client.sanitize_for_serialization(kubernetes_object)

        def drop_boring(here: Dict[str, Any]) -> None:
            """
            Drop boring fields recursively.
            """
            boring_keys = []
            for k, v in here.items():
                if isinstance(v, dict):
                    drop_boring(v)
                if k in ['managedFields']:
                    boring_keys.append(k)
            for k in boring_keys:
                del here[k]

        drop_boring(root_dict)
        return yaml.dump(root_dict)

    @overload
    def _api(
        self, kind: Literal['batch'], max_age_seconds: float = 5 * 60, errors: Optional[List[int]] = None
    ) -> BatchV1Api:
        ...

    @overload
    def _api(
        self, kind: Literal['core'], max_age_seconds: float = 5 * 60, errors: Optional[List[int]] = None
    ) -> CoreV1Api:
        ...

    @overload
    def _api(
        self, kind: Literal['customObjects'], max_age_seconds: float = 5 * 60, errors: Optional[List[int]] = None
    ) -> CustomObjectsApi:
        ...

    @overload
    def _api(
        self, kind: Literal['namespace'], max_age_seconds: float = 5 * 60
    ) -> str:
        ...

    def _api(
        self,
        kind: Union[Literal['batch'], Literal['core'], Literal['customObjects'], Literal['namespace']],
        max_age_seconds: float = 5 * 60,
        errors: Optional[List[int]] = None
    ) -> Union[BatchV1Api, CoreV1Api, CustomObjectsApi, str]:
        """
        The Kubernetes module isn't clever enough to renew its credentials when
        they are about to expire. See
        https://github.com/kubernetes-client/python/issues/741.

        We work around this by making sure that every time we are about to talk
        to Kubernetes, we have fresh credentials. And we do that by reloading
        the config and replacing our Kubernetes API objects before we do any
        Kubernetes things.

        TODO: We can still get in trouble if a single watch or listing loop
        goes on longer than our credentials last, though.

        This method is the Right Way to get any Kubernetes API. You call it
        with the API you want ('batch', 'core', or 'customObjects') and it
        returns an API object with guaranteed fresh credentials.

        It also recognizes 'namespace' and returns our namespace as a string.

        max_age_seconds needs to be << your cluster's credential expiry time.

        :param errors: If set, API calls made on the returned object will be retried
               if they produce retriable errors, except that Kubernetes API
               exceptions with HTTP status codes in the list will not be
               retried. Useful values are [] to retry on all errors, and [404]
               to retry on errors not caused by the thing in question not
               existing.
        """

        now = utc_now()

        if self.credential_time is None or (now - self.credential_time).total_seconds() > max_age_seconds:
            # Credentials need a refresh
            try:
                # Load ~/.kube/config or KUBECONFIG
                load_kube_config()
                # Worked. We're using kube config
                config_source = 'kube'
            except ConfigException:
                # Didn't work. Try pod-based credentials in case we are in a pod.
                try:
                    load_incluster_config()
                    # Worked. We're using in_cluster config
                    config_source = 'in_cluster'
                except ConfigException:
                    raise RuntimeError('Could not load Kubernetes configuration from ~/.kube/config, $KUBECONFIG, or current pod.')

            # Now fill in the API objects with these credentials
            self._apis['batch'] = BatchV1Api()
            self._apis['core'] = CoreV1Api()
            self._apis['customObjects'] = CustomObjectsApi()

            # And save the time
            self.credential_time = now
        if kind == 'namespace':
            # We just need the namespace string
            if config_source == 'in_cluster':
                # Our namespace comes from a particular file.
                with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace") as fh:
                    return fh.read().strip()
            else:
                # Find all contexts and the active context.
                # The active context gets us our namespace.
                contexts, activeContext = list_kube_config_contexts()
                if not contexts:
                    raise RuntimeError("No Kubernetes contexts available in ~/.kube/config or $KUBECONFIG")

                # Identify the namespace to work in
                namespace = activeContext.get('context', {}).get('namespace', 'default')
                assert isinstance(namespace, str)
                return namespace

        else:
            # We need an API object
            try:
                api_object = self._apis[kind]
                if errors is None:
                    # No wrapping needed
                    return api_object
                else:
                    # We need to wrap it up so that all calls that raise errors are retried.
                    error_list = list(retryable_kubernetes_errors)
                    if len(errors) > 0:
                        # Say don't retry on these particular errors
                        error_list.append(
                            ErrorCondition(
                                error=ApiException,
                                error_codes=errors,
                                retry_on_this_condition=False
                            )
                        )
                    decorator = retry(errors=error_list)
                    wrapper = KubernetesBatchSystem.DecoratorWrapper(api_object, decorator)
                    return cast(Union[BatchV1Api, CoreV1Api, CustomObjectsApi], wrapper)
            except KeyError:
                raise RuntimeError(f"Unknown Kubernetes API type: {kind}")

    class DecoratorWrapper:
        """
        Class to wrap an object so all its methods are decorated.
        """

        P = ParamSpec("P")
        def __init__(self, to_wrap: Any, decorator: Callable[[Callable[P, Any]], Callable[P, Any]]) -> None:
            """
            Make a wrapper around the given object.
            When methods on the object are called, they will be called through
            the given decorator.
            """
            self._wrapped = to_wrap
            self._decorator = decorator

        def __getattr__(self, name: str) -> Any:
            """
            Get a member as if we are actually the wrapped object.
            If it looks callable, we will decorate it.
            """

            attr = getattr(self._wrapped, name)
            if callable(attr):
                # Can be called, so return a wrapped version
                return self._decorator(attr)
            else:
                # Can't be called, so pass it by itself
                return attr

    ItemT = TypeVar("ItemT")
    class _ItemsHaver(Protocol[ItemT]):
        """
        Anything that has a .items that is a list of something.
        """
        # KubernetesBatchSystem isn't defined until the class executes, so any
        # up-references to types from there that are in signatures (and not
        # method code) need to be quoted
        items: List["KubernetesBatchSystem.ItemT"]

    CovItemT = TypeVar("CovItemT", covariant=True)
    class _WatchEvent(Protocol[CovItemT]):
        """
        An event from a Kubernetes watch stream.
        See https://github.com/kubernetes-client/python/blob/1271465acdb80bf174c50564a384fd6898635ea6/kubernetes/base/watch/watch.py#L130-L136.
        """

        # TODO: this can't be a TypedDict because of
        # https://stackoverflow.com/a/71945477 so we go with an overloaded
        # __getitem__ instead.

        @overload
        def __getitem__(self, name: Literal['type']) -> str:
            ...

        @overload
        def __getitem__(self, name: Literal['object']) -> "KubernetesBatchSystem.CovItemT":
            ...

        @overload
        def __getitem__(self, name: Literal['raw_object']) -> Dict[str, Any]:
            ...

        def __getitem__(self, name: Union[Literal['type'], Literal['object'], Literal['raw_object']]) -> Any:
            ...

    P = ParamSpec("P")
    R = TypeVar("R")
    def _stream_until_error(self, method: Callable[P, _ItemsHaver[R]], *args: P.args, **kwargs: P.kwargs) -> Iterator[_WatchEvent[R]]:
        """
        Kubernetes kubernetes.watch.Watch().stream() streams can fail and raise
        errors. We don't want to have those errors fail the entire workflow, so
        we handle them here.

        When you want to stream the results of a Kubernetes API method, call
        this instead of stream().

        To avoid having to do our own timeout logic, we finish the watch early
        if it produces an error.
        """

        w = Watch()

        # We will set this to bypass our second catch in the case of user errors.
        userError = False

        try:
            for item in w.stream(method, *args, **kwargs):
                # For everything the watch stream gives us
                try:
                    # Show the item to user code
                    yield item
                except Exception as e:
                    # If we get an error from user code, skip our catch around
                    # the Kubernetes generator.
                    userError = True
                    raise
        except Exception as e:
            # If we get an error
            if userError:
                # It wasn't from the Kubernetes watch generator. Pass it along.
                raise
            else:
                # It was from the Kubernetes watch generator we manage.
                if is_retryable_kubernetes_error(e):
                    # This is just cloud weather.
                    # TODO: We will also get an APIError if we just can't code good against Kubernetes. So make sure to warn.
                    logger.warning("Received error from Kubernetes watch stream: %s", e)
                    # Just end the watch.
                    return
                else:
                    # Something actually weird is happening.
                    raise

    def _scheduler(self) -> None:
        """
        The scheduler thread looks at jobs from the input queue, and submits
        them to the Kubernetes cluster when there are resources available.
        """
        while True:
            with self._work_available:
                # Wait until we get notified to do work.
                self._work_available.wait()

                if self._shutting_down.is_set():
                    # We're shutting down.
                    logger.info("Shutting down scheduler thread.")
                    break

                # Either we have new jobs inside the queue or more resources
                # have become available.

                # Loop through all jobs inside the queue and see if any of them
                # could be launched.
                jobs: Queue[Tuple[int, JobDescription, V1PodSpec]] = Queue()
                while True:
                    try:
                        job = self._jobs_queue.get_nowait()
                        job_id, job_desc, spec = job

                        # Check if this job has been previously killed.
                        if job_id in self._killed_queue_jobs:
                            self._killed_queue_jobs.remove(job_id)
                            logger.debug(f"Skipping killed job {job_id}")
                            continue

                        job_name = f'{self.job_prefix}{job_id}'
                        result = self._launch_job(job_name, job_desc, spec)
                        if result is False:
                            # Not enough resources to launch this job.
                            jobs.put(job)
                        else:
                            self._queued_job_ids.remove(job_id)

                    except Empty:
                        break

                # We've gone over the entire queue and these are the jobs that
                # have not been scheduled, so put them back to the queue.
                while True:
                    try:
                        self._jobs_queue.put(jobs.get_nowait())
                    except Empty:
                        break
                logger.debug(f"Roughly {self._jobs_queue.qsize} jobs in the queue")

    def setUserScript(self, userScript: Resource) -> None:
        logger.info(f'Setting user script for deployment: {userScript}')
        self.user_script = userScript

    # setEnv is provided by BatchSystemSupport, updates self.environment

    class Placement:
        """
        Internal format for pod placement constraints and preferences.
        """

        def __init__(self) -> None:
            """
            Make a new empty set of placement constraints.
            """

            self.required_labels: KeyValuesList = []
            """
            Labels which are required to be present (with these values).
            """
            self.desired_labels: KeyValuesList = []
            """
            Labels which are optional, but preferred to be present (with these values).
            """
            self.prohibited_labels: KeyValuesList = []
            """
            Labels which are not allowed to be present (with these values).
            """
            self.tolerated_taints: KeyValuesList = []
            """
            Taints which are allowed to be present (with these values).
            """

        def set_preemptible(self, preemptible: bool) -> None:
            """
            Add constraints for a job being preemptible or not.

            Preemptible jobs will be able to run on preemptible or non-preemptible
            nodes, and will prefer preemptible nodes if available.

            Non-preemptible jobs will not be allowed to run on nodes that are
            marked as preemptible.

            Understands the labeling scheme used by EKS, and the taint scheme used
            by GCE. The Toil-managed Kubernetes setup will mimic at least one of
            these.
            """

            # We consider nodes preemptible if they have any of these label or taint values.
            # We tolerate all effects of specified taints.
            # Amazon just uses a label, while Google
            # <https://cloud.google.com/kubernetes-engine/docs/how-to/preemptible-vms>
            # uses a label and a taint.
            PREEMPTIBLE_SCHEMES = {'labels': [('eks.amazonaws.com/capacityType', ['SPOT']),
                                              ('cloud.google.com/gke-preemptible', ['true'])],
                                   'taints': [('cloud.google.com/gke-preemptible', ['true'])]}

            if preemptible:
                # We want to seek preemptible labels and tolerate preemptible taints.
                self.desired_labels += PREEMPTIBLE_SCHEMES['labels']
                self.tolerated_taints += PREEMPTIBLE_SCHEMES['taints']
            else:
                # We want to prohibit preemptible labels
                self.prohibited_labels += PREEMPTIBLE_SCHEMES['labels']


        def apply(self, pod_spec: V1PodSpec) -> None:
            """
            Set ``affinity`` and/or ``tolerations`` fields on pod_spec, so that
            it runs on the right kind of nodes for the constraints we represent.
            """

            # Convert our collections to Kubernetes expressions.

            # REQUIRE that ALL of these requirements be satisfied
            required_selector_requirements: List[V1NodeSelectorRequirement] = []
            # PREFER that EACH of these terms be satisfied
            preferred_scheduling_terms: List[V1PreferredSchedulingTerm] = []
            # And this list of tolerations to apply
            tolerations: List[V1Toleration] = []

            for label, values in self.required_labels:
                # Collect requirements for the required labels
                has_label = V1NodeSelectorRequirement(key=label,
                                                      operator='In',
                                                      values=values)
                required_selector_requirements.append(has_label)
            for label, values in self.desired_labels:
                # Collect preferences for the preferred labels
                has_label = V1NodeSelectorRequirement(key=label,
                                                      operator='In',
                                                      values=values)
                term = V1NodeSelectorTerm(
                    match_expressions=[has_label]
                )
                # Each becomes a separate preference, more is better.
                preference = V1PreferredSchedulingTerm(weight=1,
                                                       preference=term)

                preferred_scheduling_terms.append(preference)
            for label, values in self.prohibited_labels:
                # So we need to say that each label either doesn't
                # have any of the banned values, or doesn't exist.
                # Although the docs don't really say so, NotIn also matches
                # cases where the label doesn't exist. This is suggested by
                # <https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#set-based-requirement>
                # So we create a NotIn for each label and AND them
                # all together.
                not_labeled = V1NodeSelectorRequirement(key=label,
                                                        operator='NotIn',
                                                        values=values)
                required_selector_requirements.append(not_labeled)
            for taint, values in self.tolerated_taints:
                for value in values:
                    # Each toleration can tolerate one value
                    taint_ok = V1Toleration(key=taint,
                                            value=value)
                    tolerations.append(taint_ok)

            # Now combine everything
            if preferred_scheduling_terms or required_selector_requirements:
                # We prefer or require something about labels.

                requirements_selector: Optional[V1NodeSelector] = None
                if required_selector_requirements:
                    # Make a term that says we match all the requirements
                    requirements_term = V1NodeSelectorTerm(
                        match_expressions=required_selector_requirements
                    )
                    # And a selector to hold the term
                    requirements_selector = V1NodeSelector(node_selector_terms=[requirements_term])

                # Make an affinity that prefers the preferences and requires the requirements
                node_affinity = V1NodeAffinity(
                    preferred_during_scheduling_ignored_during_execution=preferred_scheduling_terms if preferred_scheduling_terms else None,
                    required_during_scheduling_ignored_during_execution=requirements_selector
                )

                # Apply the affinity
                pod_spec.affinity = V1Affinity(node_affinity = node_affinity)

            if tolerations:
                # Apply the tolerations
                pod_spec.tolerations = tolerations

    def _check_accelerator_request(self, requirer: Requirer) -> None:
        for accelerator in requirer.accelerators:
            if accelerator['kind'] != 'gpu' and 'model' not in accelerator:
                # We can only provide GPUs or things with a model right now
                raise InsufficientSystemResources(requirer, 'accelerators', details=[
                    f'The accelerator {accelerator} could not be provided.',
                    'The Toil Kubernetes batch system only knows how to request gpu accelerators or accelerators with a defined model.'
                ])

    def _create_pod_spec(
            self,
            job_desc: JobDescription,
            job_environment: Optional[Dict[str, str]] = None
    ) -> V1PodSpec:
        """
        Make the specification for a pod that can execute the given job.
        """

        environment = self.environment.copy()
        if job_environment:
            environment.update(job_environment)

        # Make a command to run it in the executor
        command_list = pack_job(job_desc, self.user_script, environment=environment)

        # The Kubernetes API makes sense only in terms of the YAML format. Objects
        # represent sections of the YAML files. Except from our point of view, all
        # the internal nodes in the YAML structure are named and typed.

        # For docs, start at the root of the job hierarchy:
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Job.md

        # Make a definition for the container's resource requirements.
        # Add on a bit for Kubernetes overhead (Toil worker's memory, hot deployed
        # user scripts).
        # Kubernetes needs some lower limit of memory to run the pod at all without
        # OOMing. We also want to provision some extra space so that when
        # we test _isPodStuckOOM we never get True unless the job has
        # exceeded job_desc.memory.
        requirements_dict = {'cpu': job_desc.cores,
                             'memory': job_desc.memory + 1024 * 1024 * 512,
                             'ephemeral-storage': job_desc.disk + 1024 * 1024 * 512}

        # Also start on the placement constraints
        placement = KubernetesBatchSystem.Placement()
        placement.set_preemptible(job_desc.preemptible)

        for accelerator in job_desc.accelerators:
            # Add in requirements for accelerators (GPUs).
            # See https://kubernetes.io/docs/tasks/manage-gpus/scheduling-gpus/

            if accelerator['kind'] == 'gpu':
                # We can't schedule GPUs without a brand, because the
                # Kubernetes resources are <brand>.com/gpu. If no brand is
                # specified, default to nvidia, which is very popular.
                vendor = accelerator.get('brand', 'nvidia')
                key = f'{vendor}.com/{accelerator["kind"]}'
                if key not in requirements_dict:
                    requirements_dict[key] = 0
                requirements_dict[key] += accelerator['count']

            if 'model' in accelerator:
                # TODO: What if the cluster uses some other accelerator model labeling scheme?
                placement.required_labels.append(('accelerator', [accelerator['model']]))

            # TODO: Support AMD's labeling scheme: https://github.com/RadeonOpenCompute/k8s-device-plugin/tree/master/cmd/k8s-node-labeller
            # That just has each trait of the accelerator as a separate label, but nothing that quite corresponds to a model.

        # Make all the resource requests into strings
        requests_dict = {k: str(v) for k, v in requirements_dict.items()}

        # Use the requests as the limits, for predictable behavior, and because
        # the UCSC Kubernetes admins want it that way. For GPUs, Kubernetes
        # requires them to be equal.
        limits_dict = requests_dict
        resources = V1ResourceRequirements(limits=limits_dict,
                                           requests=requests_dict)

        # Collect volumes and mounts
        volumes = []
        mounts = []

        def mount_host_path(volume_name: str, host_path: str, mount_path: str, create: bool = False) -> None:
            """
            Add a host path volume with the given name to mount the given path.

            :param create: If True, create the directory on the host if it does
                   not exist. Otherwise, when the directory does not exist, the
                   pod will wait for it to come into existence.
            """
            volume_type = 'DirectoryOrCreate' if create else 'Directory'
            volume_source = V1HostPathVolumeSource(path=host_path, type=volume_type)
            volume = V1Volume(name=volume_name,
                                                host_path=volume_source)
            volumes.append(volume)
            volume_mount = V1VolumeMount(mount_path=mount_path, name=volume_name)
            mounts.append(volume_mount)

        if self.host_path is not None:
            # Provision Toil WorkDir from a HostPath volume, to share with other pods.
            # Create the directory if it doesn't exist already.
            mount_host_path('workdir', self.host_path, self.worker_work_dir, create=True)
            # We also need to mount across /run/lock, where we will put
            # per-node coordiantion info.
            # Don't create this; it really should always exist.
            mount_host_path('coordination', '/run/lock', '/run/lock')
        else:
            # Provision Toil WorkDir as an ephemeral volume
            ephemeral_volume_name = 'workdir'
            ephemeral_volume_source = V1EmptyDirVolumeSource()
            ephemeral_volume = V1Volume(name=ephemeral_volume_name,
                                                          empty_dir=ephemeral_volume_source)
            volumes.append(ephemeral_volume)
            ephemeral_volume_mount = V1VolumeMount(mount_path=self.worker_work_dir, name=ephemeral_volume_name)
            mounts.append(ephemeral_volume_mount)
            # And don't share coordination directory

        if self.aws_secret_name is not None:
            # Also mount an AWS secret, if provided.
            # TODO: make this generic somehow
            secret_volume_name = 's3-credentials'
            secret_volume_source = V1SecretVolumeSource(secret_name=self.aws_secret_name)
            secret_volume = V1Volume(name=secret_volume_name,
                                                       secret=secret_volume_source)
            volumes.append(secret_volume)
            secret_volume_mount = V1VolumeMount(mount_path='/root/.aws', name=secret_volume_name)
            mounts.append(secret_volume_mount)

        # Make a container definition
        container = V1Container(command=command_list,
                                                  image=self.docker_image,
                                                  name="runner-container",
                                                  resources=resources,
                                                  volume_mounts=mounts)
        # Wrap the container in a spec
        pod_spec = V1PodSpec(containers=[container],
                                               volumes=volumes,
                                               restart_policy="Never")
        # Tell the spec where to land
        placement.apply(pod_spec)

        if self.service_account:
            # Apply service account if set
            pod_spec.service_account_name = self.service_account

        return pod_spec

    def _release_acquired_resources(self, resources: List[int], notify: bool = False) -> None:
        """
        Release all resources acquired for a job.

        :param resources: The resources in the order: core, memory, disk.
        :param notify: If True, notify the threads that are waiting on the
                       self._work_available condition.
        """
        # TODO: accelerators?

        for resource, request in zip(self.resource_sources, resources):
            assert isinstance(resource, ResourcePool) and isinstance(request, int)
            resource.release(request)

        if notify:
            with self._work_available:
                self._work_available.notify_all()

    def _launch_job(
        self,
        job_name: str,
        job_desc: JobDescription,
        pod_spec: V1PodSpec
    ) -> bool:
        """
        Try to launch the given job to the Kubernetes cluster. Return False if
        we don't have enough resources to submit the job.
        """

        # Limit the amount of resources requested at a time.
        resource_requests: List[int] = [1, int(job_desc.cores * 1000), job_desc.memory, job_desc.disk]

        acquired = []
        for source, request in zip(self.resource_sources, resource_requests):
            # For each kind of resource we want, go get it
            assert ((isinstance(source, ResourcePool) and isinstance(request, int)))
            if source.acquireNow(request):
                acquired.append(request)
            else:
                # We can't get everything
                self._release_acquired_resources(acquired,
                    # Put it back quietly.
                    notify=False)
                return False

        self._acquired_resources[job_name] = acquired

        # We have all the resources we need; submit it to the cluster.

        # Make metadata to label the job/pod with info.
        # Don't let the cluster autoscaler evict any Toil jobs.
        metadata = V1ObjectMeta(name=job_name,
                                labels={"toil_run": self.run_id},
                                annotations={"cluster-autoscaler.kubernetes.io/safe-to-evict": "false"})

        # Wrap the spec in a template
        template = V1PodTemplateSpec(spec=pod_spec, metadata=metadata)

        # Make another spec for the job, asking to run the template with no
        # backoff/retry. Specify our own TTL to avoid catching the notice
        # of over-zealous abandoned job cleanup scripts.
        job_spec = V1JobSpec(template=template,
                             backoff_limit=0,
                             ttl_seconds_after_finished=self.finished_job_ttl)

        # And make the actual job
        job = V1Job(spec=job_spec,
                    metadata=metadata,
                    api_version="batch/v1",
                    kind="Job")

        # Launch the job
        launched = self._api('batch', errors=[]).create_namespaced_job(self.namespace, job)

        logger.debug(f"Launched job: {job_name}")

        return True

    def _delete_job(
        self,
        job_name: str, *,
        propagation_policy: Literal["Foreground", "Background"] = "Foreground",
        gone_ok: bool = False,
        resource_notify: bool = True
    ) -> None:
        """
        Given the name of a kubernetes job, delete the job and release all
        acquired resources for the job.

        :param job_name: The Kubernetes job name.
        :param propagation_policy: The propagation policy for the delete.
        :param gone_ok: If True, don't raise an error if the job is already gone.
        :param resource_notify: If True, notify the threads that are waiting on
            the self._work_available condition.
        """
        try:
            logger.debug(f'Deleting Kubernetes job {job_name}')
            self._api('batch', errors=[404] if gone_ok else []).delete_namespaced_job(
                job_name,
                self.namespace,
                propagation_policy=propagation_policy
            )
        finally:
            # We should always release the acquired resources.
            resources = self._acquired_resources.get(job_name)
            if not resources:
                logger.warning(f"Cannot get the acquired resources from {job_name}")
                return
            self._release_acquired_resources(resources, notify=resource_notify)
            del self._acquired_resources[job_name]

    def issueBatchJob(self, job_desc: JobDescription, job_environment: Optional[Dict[str, str]] = None) -> int:
        # Try the job as local
        localID = self.handleLocalJob(job_desc)
        if localID is not None:
            # It is a local job
            return localID

        # We actually want to send to the cluster

        # Check resource requirements
        self.check_resource_request(job_desc)

        # Make a pod that describes running the job
        pod_spec = self._create_pod_spec(job_desc, job_environment=job_environment)

        # Make a batch system scope job ID
        job_id = self.getNextJobID()

        # Put job inside queue to be launched to the cluster
        self._queued_job_ids.add(job_id)

        # Also keep track of the ID for getIssuedBatchJobIDs()
        self._jobs_queue.put((job_id, job_desc, pod_spec))

        # Let scheduler know that there is work to do.
        with self._work_available:
            self._work_available.notify_all()

        logger.debug(f"Enqueued job {job_id}")

        return job_id

    class _ArgsDict(TypedDict):
        """
        Dict for kwargs to list_namespaced_job/pod, so MyPy can know we're
        passign the right types when we make a dict and unpack it into kwargs.

        We need to do unpacking because the methods inspect the keys in their
        kwargs, so we can't just set unused ones to None. But we also don't
        want to duplicate code for every combination of possible present keys.
        """
        _continue: NotRequired[str]
        label_selector: NotRequired[str]
        field_selector: NotRequired[str]

    def _ourJobObject(self, onlySucceeded: bool = False) -> Iterator[V1Job]:
        """
        Yield Kubernetes V1Job objects that we are responsible for that the
        cluster knows about. Ignores jobs in the process of being deleted.

        Doesn't support a free-form selector, because there's only about 3
        things jobs can be selected on: https://stackoverflow.com/a/55808444

        :param bool onlySucceeded: restrict results to succeeded jobs.
        :param int limit: max results to yield.
        """

        # We need to page through the list from the cluster with a continuation
        # token. These expire after about 5 minutes. If we use an expired one,
        # we get a 410 error and a new token, and we can use the new token to
        # get the rest of the list, but the list will be updated.
        #
        # TODO: How to get the new token isn't clear. See
        # https://github.com/kubernetes-client/python/issues/953. For now we
        # will just throw an error if we don't get to the end of the list in
        # time.

        token = None

        while True:
            kwargs: KubernetesBatchSystem._ArgsDict = {'label_selector': f"toil_run={self.run_id}"}

            if onlySucceeded:
                kwargs['field_selector'] = "status.successful==1"

            if token is not None:
                kwargs['_continue'] = token

            results = self._api('batch', errors=[]).list_namespaced_job(
                self.namespace,
                **kwargs
            )
            
            # These jobs belong to us
            yield from (j for j in results.items if not self._is_deleted(j))

            # Remember the continuation token, if any
            token = getattr(results.metadata, 'continue', None)

            if token is None:
                # There isn't one. We got everything.
                break


    def _ourPodObject(self) -> Iterator[V1Pod]:
        """
        Yield Kubernetes V1Pod objects that we are responsible for that the
        cluster knows about.
        """

        token = None

        while True:
            kwargs: KubernetesBatchSystem._ArgsDict = {'label_selector': f"toil_run={self.run_id}"}

            if token is not None:
                kwargs['_continue'] = token

            results = self._api('core', errors=[]).list_namespaced_pod(
                self.namespace,
                **kwargs
            )

            yield from (j for j in results.items if not self._is_deleted(j))
            # Remember the continuation token, if any
            token = getattr(results.metadata, 'continue', None)

            if token is None:
                # There isn't one. We got everything.
                break


    def _getPodForJob(self, jobObject: V1Job) -> Optional[V1Pod]:
        """
        Get the pod that belongs to the given job, or None if the job's pod is
        missing. The pod knows about things like the job's exit code.

        :param V1Job jobObject: a Kubernetes job to look up
                                       pods for.

        :return: The pod for the job, or None if no pod is found.
        :rtype: V1Pod
        """

        # Make sure the job has the fields we need
        assert(jobObject.metadata is not None)

        token = None

        while True:
            kwargs: KubernetesBatchSystem._ArgsDict = {'label_selector': f'job-name={jobObject.metadata.name}'}
            if token is not None:
                kwargs['_continue'] = token
            results = self._api('core', errors=[]).list_namespaced_pod(self.namespace, **kwargs)

            for pod in results.items:
                # Return the first pod we find
                return pod

            # Remember the continuation token, if any
            token = getattr(results.metadata, 'continue', None)

            if token is None:
                # There isn't one. We got everything.
                break

        # If we get here, no pages had any pods.
        return None

    def _getLogForPod(self, podObject: V1Pod) -> str:
        """
        Get the log for a pod.

        :param V1Pod podObject: a Kubernetes pod with one
               container to get the log from.

        :return: The log for the only container in the pod.
        :rtype: str

        """

        assert podObject.metadata is not None
        assert podObject.metadata.name is not None

        return self._api('core', errors=[]).read_namespaced_pod_log(
            podObject.metadata.name,
            namespace=self.namespace
        )

    def _isPodStuckOOM(self, podObject: V1Pod, minFreeBytes: float = 1024 * 1024 * 2) -> bool:
        """
        Poll the current memory usage for the pod from the cluster.

        Return True if the pod looks to be in a soft/stuck out of memory (OOM)
        state, where it is using too much memory to actually make progress, but
        not enough to actually trigger the OOM killer to kill it. For some
        large memory limits, on some Kubernetes clusters, pods can get stuck in
        this state when their memory limits are high (approx. 200 Gi).

        We operationalize "OOM" as having fewer than minFreeBytes bytes free.

        We assume the pod has only one container, as Toil's pods do.

        If the metrics service is not working, we treat the pod as not being
        stuck OOM. Otherwise, we would kill all functioning jobs on clusters
        where the metrics service is down or isn't installed.

        :param V1Pod podObject: a Kubernetes pod with one
                                       container to check up on.
        :param int minFreeBytes: Minimum free bytes to not be OOM.

        :return: True if the pod is OOM, False otherwise.
        :rtype: bool
        """

        assert podObject.metadata is not None
        assert podObject.metadata.name is not None

        # Compose a query to get just the pod we care about
        query = f'metadata.name={podObject.metadata.name}'

        # Look for it, but manage our own exceptions
        try:
            # TODO: When the Kubernetes Python API actually wraps the metrics API, switch to that
            response = self._api('customObjects').list_namespaced_custom_object('metrics.k8s.io', 'v1beta1',
                                                                                self.namespace, 'pods',
                                                                                field_selector=query)
        except Exception as e:
            # We couldn't talk to the metrics service on this attempt. We don't
            # retry, but we also don't want to just ignore all errors. We only
            # want to ignore errors we expect to see if the problem is that the
            # metrics service is not working.
            if type(e) in retryable_kubernetes_errors:
                # This is the sort of error we would expect from an overloaded
                # Kubernetes or a dead metrics service.
                # We can't tell that the pod is stuck, so say that it isn't.
                logger.warning("Could not query metrics service: %s", e)
                return False
            else:
                raise

        # Pull out the items
        items = response.get('items', [])

        if len(items) == 0:
            # If there's no statistics we can't say we're stuck OOM
            return False

        # Assume the first result is the right one, because of the selector.
        # That means we don't need to bother with _continue.
        # Assume it has exactly one pod, because we made it.
        containers = items[0].get('containers', [{}])

        if len(containers) == 0:
            # If there are no containers (because none have started yet?), we can't say we're stuck OOM
            return False

        # Otherwise, assume it just has one container.
        # Grab the memory usage string, like 123Ki, and convert to bytes.
        # If anything is missing, assume 0 bytes used.
        bytesUsed = human2bytes(containers[0].get('usage', {}).get('memory', '0'))

        # Also get the limit out of the pod object's spec
        assert podObject.spec is not None
        assert len(podObject.spec.containers) > 0
        assert podObject.spec.containers[0].resources is not None
        assert podObject.spec.containers[0].resources.limits is not None
        assert 'memory' in podObject.spec.containers[0].resources.limits
        bytesAllowed = human2bytes(podObject.spec.containers[0].resources.limits['memory'])

        if bytesAllowed - bytesUsed < minFreeBytes:
            # This is too much!
            logger.warning('Pod %s has used %d of %d bytes of memory; reporting as stuck due to OOM.',
                           podObject.metadata.name, bytesUsed, bytesAllowed)

            return True
        else:
            return False

    def _isPodStuckWaiting(self, pod_object: V1Pod, reason: Optional[str] = None, timeout: Optional[float] = None) -> bool:
        """
        Return True if the pod looks to be in a waiting state, and false otherwise.

        :param pod_object: a Kubernetes pod with one container to check up on.
        :param reason: Only match pods with a waiting reason equal to this.
        :param timeout: Only match pods that have a start_time (scheduling time) older than this many seconds.

        :return: True if the pod is probably stuck, False otherwise.
        """

        if pod_object.status is None:
            return False

        # Get the statuses of the pod's containers
        container_statuses = pod_object.status.container_statuses
        if container_statuses is None or len(container_statuses) == 0:
            # Pod exists but has no container statuses
            # This happens when the pod is just "Scheduled"
            # ("PodScheduled" status event) and isn't actually starting
            # to run yet.
            # Can't be stuck
            return False

        waiting_info = getattr(getattr(container_statuses[0], 'state', None), 'waiting', None)
        if waiting_info is None:
            # Pod is not waiting
            return False

        if reason is not None and waiting_info.reason != reason:
            # Pod fails reason filter
            return False

        start_time = getattr(pod_object.status, 'start_time', None)
        if timeout is not None and (start_time is None or (utc_now() - start_time).total_seconds() < timeout):
            # It hasn't been waiting too long, or we care but don't know how
            # long it has been waiting
            return False

        return True

    def _is_deleted(self, kube_thing: Union['V1Job', 'V1Pod']) -> bool:
        """
        Determine if a job or pod is in the process od being deleted, and
        shouldn't count anymore.
        """

        # Kubernetes "Terminating" is the same as having the deletion_timestamp
        # set in the metadata of the object.

        deletion_timestamp: Optional[datetime.datetime] = getattr(getattr(kube_thing, 'metadata', None), 'deletion_timestamp', None)
        # If the deletion timestamp is set to anything, it is in the process of
        # being deleted. We will treat that as as good as gone.
        return deletion_timestamp is not None

    def _getIDForOurJob(self, jobObject: V1Job) -> int:
        """
        Get the JobID number that belongs to the given job that we own.

        :param V1Job jobObject: a Kubernetes job object that is a job we issued.

        :return: The JobID for the job.
        :rtype: int
        """

        assert jobObject.metadata is not None
        assert jobObject.metadata.name is not None
        return int(jobObject.metadata.name[len(self.job_prefix):])


    def getUpdatedBatchJob(self, maxWait: float) -> Optional[UpdatedBatchJobInfo]:

        entry = datetime.datetime.now()

        result = self._getUpdatedBatchJobImmediately()

        if result is not None or maxWait == 0:
            # We got something on the first try, or we only get one try.
            # If we wait
            return result

        # Otherwise we need to maybe wait.
        if self.enable_watching and maxWait >= 1:
            # We can try a watch. Watches can only work in whole seconds.
            for event in self._stream_until_error(self._api('batch').list_namespaced_job,
                                                  self.namespace,
                                                  label_selector=f"toil_run={self.run_id}",
                                                  timeout_seconds=math.floor(maxWait)):
                # Grab the metadata data, ID, the list of conditions of the current job, and the total pods
                jobObject = event['object']
                
                if self._is_deleted(jobObject):
                    # Job is already deleted, so ignore it.
                    logger.warning('Kubernetes job %s is deleted; ignore its update', getattr(getattr(jobObject, 'metadata', None), 'name', None))
                    continue
                
                assert jobObject.metadata is not None
                assert jobObject.metadata.name is not None
                
                jobID = int(jobObject.metadata.name[len(self.job_prefix):])
                if jobObject.status is None:
                    # Can't tell what is up with this job.
                    continue
                # Fetch out the pod counts, make sure they are numbers
                active_pods = jobObject.status.active or 0
                succeeded_pods = jobObject.status.succeeded or 0
                failed_pods = jobObject.status.failed or 0
                # Fetch out the condition object that has info about how the job is going.
                condition: Optional[V1JobCondition] = None
                if jobObject.status.conditions is not None and len(jobObject.status.conditions) > 0:
                    condition = jobObject.status.conditions[0]

                totalPods = active_pods + succeeded_pods + failed_pods
                # Exit Reason defaults to 'Successfully Finished` unless said otherwise
                exitReason = BatchJobExitReason.FINISHED
                exitCode = 0

                # Check if there are any active pods
                if active_pods > 0:
                    logger.info("%s has %d pods running" % jobObject.metadata.name, active_pods)
                    continue
                elif succeeded_pods > 0 or failed_pods > 0:
                    # No more active pods in the current job ; must be finished
                    logger.info("%s RESULTS -> Succeeded: %d Failed:%d Active:%d" % jobObject.metadata.name,
                                                                                    succeeded_pods, failed_pods, active_pods)
                    # Log out success/failure given a reason
                    logger.info("%s REASON: %s", getattr(condition, 'type', None), getattr(condition, 'reason', None))

                    # Log out reason of failure and pod exit code
                    if failed_pods > 0:
                        exitReason = BatchJobExitReason.FAILED
                        exitCode = EXIT_STATUS_UNAVAILABLE_VALUE
                        logger.debug("Failed job %s", self._pretty_print(jobObject))
                        if condition is not None:
                            logger.warning("Failed Job Message: %s", condition.message)
                        pod = self._getPodForJob(jobObject)
                        statuses: List[V1ContainerStatus] = getattr(getattr(pod, 'status', None), 'container_statuses', [])
                        if len(statuses) > 0 and statuses[0].state is not None and statuses[0].state.terminated is not None:
                            exitCode = statuses[0].state.terminated.exit_code

                    raw_runtime = 0.0
                    if jobObject.status.completion_time is not None and jobObject.status.start_time is not None:
                        raw_runtime = (jobObject.status.completion_time - jobObject.status.start_time).total_seconds()
                    runtime = slow_down(raw_runtime)
                    result = UpdatedBatchJobInfo(jobID=jobID, exitStatus=exitCode, wallTime=runtime, exitReason=exitReason)

                    if (exitReason == BatchJobExitReason.FAILED) or (succeeded_pods + failed_pods == totalPods):
                        # Cleanup if job is all finished or there was a pod that failed
                        # TODO: use delete_job() to release acquired resources
                        self._delete_job(
                            jobObject.metadata.name,
                            propagation_policy='Foreground'
                        )
                        # Make sure the job is deleted so we won't see it again.
                        self._waitForJobDeath(jobObject.metadata.name)
                        return result
                    continue
                else:
                    # Job is not running/updating ; no active, successful, or failed pods yet
                    logger.debug("Job {} -> {}".format(jobObject.metadata.name, getattr(condition, 'reason', None)))
                    # Pod could be pending; don't say it's lost.
                    continue
        else:
            # Try polling instead
            while result is None and (datetime.datetime.now() - entry).total_seconds() < maxWait:
                # We still have nothing and we haven't hit the timeout.

                # Poll
                result = self._getUpdatedBatchJobImmediately()

                if result is None:
                    # Still nothing. Wait a second, or some fraction of our max wait time.
                    time.sleep(min(maxWait/2, 1.0))

            # When we get here, either we found something or we ran out of time
            return result


    def _getUpdatedBatchJobImmediately(self) -> Optional[UpdatedBatchJobInfo]:
        """
        Return None if no updated (completed or failed) batch job is currently
        available, and jobID, exitCode, runtime if such a job can be found.
        """

        # See if a local batch job has updated and is available immediately
        local_tuple = self.getUpdatedLocalJob(0)
        if local_tuple:
            # If so, use it
            return local_tuple

        # Otherwise we didn't get a local job.

        # Go looking for other jobs

        # Everybody else does this with a queue and some other thread that
        # is responsible for populating it.
        # But we can just ask kubernetes now.

        # Find a job that is done, failed, or stuck
        jobObject = None
        # Put 'done', 'failed', or 'stuck' here
        chosenFor = ''

        for j in self._ourJobObject(onlySucceeded=True):
            # Look for succeeded jobs because that's the only filter Kubernetes has
            jobObject = j
            chosenFor = 'done'

        if jobObject is None:
            for j in self._ourJobObject():
                # If there aren't any succeeded jobs, scan all jobs
                # See how many times each failed
                failCount = getattr(j.status, 'failed', 0)
                if failCount is None:
                    # Make sure it is an int
                    failCount = 0
                if failCount > 0:
                    # Take the first failed one you find
                    jobObject = j
                    chosenFor = 'failed'
                    break

        if jobObject is None:
            # If no jobs are failed, look for jobs with pods that are stuck for various reasons.
            for j in self._ourJobObject():
                pod = self._getPodForJob(j)

                if pod is None:
                    # Skip jobs with no pod
                    continue

                # Containers can get stuck in Waiting with reason ImagePullBackOff
                if self._isPodStuckWaiting(pod, reason='ImagePullBackoff'):
                    # Assume it will never finish, even if the registry comes back or whatever.
                    # We can get into this state when we send in a non-existent image.
                    # See https://github.com/kubernetes/kubernetes/issues/58384
                    jobObject = j
                    chosenFor = 'stuck'
                    logger.warning('Failing stuck job (ImagePullBackoff); did you try to run a non-existent Docker image?'
                                   ' Check TOIL_APPLIANCE_SELF.')
                    break

                # Containers can also get stuck in Waiting with reason
                # ContainerCreating, if for example their mounts don't work.
                if self._isPodStuckWaiting(pod, reason='ContainerCreating', timeout=self.pod_timeout):
                    # Assume that it will never finish.
                    jobObject = j
                    chosenFor = 'stuck'
                    logger.warning('Failing stuck job (ContainerCreating longer than %s seconds); did you try to mount something impossible?', self.pod_timeout)
                    break

                # Pods can also get stuck nearly but not quite out of memory,
                # if their memory limits are high and they try to exhaust them.

                if self._isPodStuckOOM(pod):
                    # We found a job that probably should be OOM! Report it as stuck.
                    # Polling function takes care of the logging.
                    jobObject = j
                    chosenFor = 'stuck'
                    break

        if jobObject is None:
            # Say we couldn't find anything
            return None
        else:
            # We actually have something
            logger.debug('Identified stopped Kubernetes job %s as %s', getattr(jobObject.metadata, 'name', None), chosenFor)


        # Otherwise we got something.

        # Work out what the job's ID was (whatever came after our name prefix)
        assert jobObject.metadata is not None
        assert jobObject.metadata.name is not None
        jobID = int(jobObject.metadata.name[len(self.job_prefix):])

        # Grab the pod
        pod = self._getPodForJob(jobObject)

        if pod is not None:
            if chosenFor == 'done' or chosenFor == 'failed':
                # The job actually finished or failed

                # Get the statuses of the pod's containers
                containerStatuses = getattr(getattr(pod, 'status', None), 'container_statuses', None)

                # Get when the pod started (reached the Kubelet) as a datetime
                start_time = self._get_start_time(pod, jobObject)

                if containerStatuses is None or len(containerStatuses) == 0:
                    # No statuses available.
                    # This happens when a pod is "Scheduled". But how could a
                    # 'done' or 'failed' pod be merely "Scheduled"?
                    # Complain so we can find out.
                    logger.warning('Exit code and runtime unavailable; pod has no container statuses')
                    logger.warning('Pod: %s', str(pod))
                    exitCode = EXIT_STATUS_UNAVAILABLE_VALUE
                    # Say it stopped now and started when it was scheduled/submitted.
                    # We still need a strictly positive runtime.
                    runtime = slow_down((utc_now() - start_time).total_seconds())
                else:
                    # Get the termination info from the pod's main (only) container
                    terminatedInfo = getattr(getattr(containerStatuses[0], 'state', None), 'terminated', None)
                    if terminatedInfo is None:
                        logger.warning('Exit code and runtime unavailable; pod stopped without container terminating')
                        logger.warning('Pod: %s', str(pod))
                        exitCode = EXIT_STATUS_UNAVAILABLE_VALUE
                        # Say it stopped now and started when it was scheduled/submitted.
                        # We still need a strictly positive runtime.
                        runtime = slow_down((utc_now() - start_time).total_seconds())
                    else:
                        # Extract the exit code
                        exitCode = terminatedInfo.exit_code

                        # Compute how long the job actually ran for (subtract
                        # datetimes). We need to look at the pod's start time
                        # because the job's start time is just when the job is
                        # created. And we need to look at the pod's end time
                        # because the job only gets a completion time if
                        # successful.
                        runtime = slow_down((terminatedInfo.finished_at -
                                             start_time).total_seconds())

                        if chosenFor == 'failed':
                            # Warn the user with the failed pod's log
                            # TODO: cut this down somehow?
                            logger.warning('Log from failed pod: %s', self._getLogForPod(pod))

            else:
                # The job has gotten stuck

                assert chosenFor == 'stuck'

                # Synthesize an exit code
                exitCode = EXIT_STATUS_UNAVAILABLE_VALUE
                # Say it ran from when the job was submitted to when the pod got stuck
                runtime = slow_down((utc_now() - self._get_start_time(job=jobObject)).total_seconds())
        else:
            # The pod went away from under the job.
            logging.warning('Exit code and runtime unavailable; pod vanished')
            exitCode = EXIT_STATUS_UNAVAILABLE_VALUE
            # Say it ran from when the job was submitted to when the pod vanished
            runtime = slow_down((utc_now() - self._get_start_time(job=jobObject)).total_seconds())


        try:
            # Delete the job and all dependents (pods), hoping to get a 404 if it's magically gone
            self._delete_job(jobObject.metadata.name, propagation_policy='Foreground', gone_ok=True)

            # That just kicks off the deletion process. Foreground doesn't
            # actually block. See
            # https://kubernetes.io/docs/concepts/workloads/controllers/garbage-collection/#foreground-cascading-deletion
            # We wait for the job to vanish, or at least pass _is_deleted so
            # we ignore it later.
            self._waitForJobDeath(jobObject.metadata.name)

        except ApiException as e:
            if e.status != 404:
                # Something is wrong, other than the job already being deleted.
                raise
            # Otherwise everything is fine and the job is gone.

        # Return the one finished job we found
        return UpdatedBatchJobInfo(jobID=jobID, exitStatus=exitCode, wallTime=runtime, exitReason=None)

    def _waitForJobDeath(self, jobName: str) -> None:
        """
        Block until the job with the given name no longer exists.
        """

        # We do some exponential backoff on the polling
        # TODO: use a wait instead of polling?
        backoffTime = 0.1
        maxBackoffTime = 6.4
        while True:
            try:
                # Look for the job
                job_object = self._api('batch', errors=[404]).read_namespaced_job(jobName, self.namespace)
                if self._is_deleted(job_object):
                    # The job looks deleted, so we can treat it as not being there.
                    return
                # If we didn't 404, and the job doesn't look deleted, wait a
                # bit with exponential backoff
                time.sleep(backoffTime)
                if backoffTime < maxBackoffTime:
                    backoffTime *= 2
            except ApiException as e:
                # We finally got a failure!
                if e.status != 404:
                    # But it wasn't due to the job being gone; something is wrong.
                    raise
                # It was a 404; the job is gone. Stop polling it.
                break

    def shutdown(self) -> None:

        # Shutdown local processes first
        self.shutdownLocal()

        # Shutdown scheduling thread
        self._shutting_down.set()
        with self._work_available:
            self._work_available.notify_all()   # Wake it up.

        self.schedulingThread.join()

        # Kill all of our jobs and clean up pods that are associated with those jobs
        try:
            logger.debug('Deleting all Kubernetes jobs for toil_run=%s', self.run_id)
            self._api('batch', errors=[404]).delete_collection_namespaced_job(
                self.namespace,
                label_selector=f"toil_run={self.run_id}",
                propagation_policy='Background'
            )
            logger.debug('Killed jobs with delete_collection_namespaced_job; cleaned up')
            # TODO: should we release all resources? We're shutting down so would it matter?
        except ApiException as e:
            if e.status != 404:
                # Anything other than a 404 is weird here.
                logger.error("Exception when calling BatchV1Api->delete_collection_namespaced_job: %s" % e)

            # If batch delete fails, try to delete all remaining jobs individually.
            logger.debug('Deleting Kubernetes jobs individually for toil_run=%s', self.run_id)
            for job_id in self._getIssuedNonLocalBatchJobIDs():
                job_name = f'{self.job_prefix}{job_id}'
                self._delete_job(job_name, propagation_policy='Background', resource_notify=False)

            # Aggregate all pods and check if any pod has failed to cleanup or is orphaned.
            ourPods = self._ourPodObject()

            for pod in ourPods:
                try:
                    phase = getattr(pod.status, 'phase', None)
                    if phase == 'Failed':
                            logger.debug('Failed pod encountered at shutdown:\n%s', self._pretty_print(pod))
                    if phase == 'Orphaned':
                            logger.debug('Orphaned pod encountered at shutdown:\n%s', self._pretty_print(pod))
                except:
                    # Don't get mad if that doesn't work.
                    pass
                if pod.metadata is not None and pod.metadata.name is not None:
                    try:
                        logger.debug('Cleaning up pod at shutdown: %s', pod.metadata.name)
                        response = self._api('core', errors=[404]).delete_namespaced_pod(
                            pod.metadata.name,
                            self.namespace,
                            propagation_policy='Background'
                        )
                    except ApiException as e:
                        if e.status != 404:
                            # Anything other than a 404 is weird here.
                            logger.error("Exception when calling CoreV1Api->delete_namespaced_pod: %s" % e)


    def _getIssuedNonLocalBatchJobIDs(self) -> List[int]:
        """
        Get the issued batch job IDs that are not for local jobs.
        """
        jobIDs = []
        with self._mutex:
            got_list = self._ourJobObject()
            for job in got_list:
                # Get the ID for each job
                jobIDs.append(self._getIDForOurJob(job))
        return jobIDs

    def getIssuedBatchJobIDs(self) -> List[int]:
        # Make sure to send the local jobs and queued jobs also
        with self._mutex:
            queued_jobs = list(self._queued_job_ids)
        return self._getIssuedNonLocalBatchJobIDs() + list(self.getIssuedLocalJobIDs()) + queued_jobs

    def _get_start_time(self, pod: Optional[V1Pod] = None, job: Optional[V1Job] = None) -> datetime.datetime:
        """
        Get an actual or estimated start time for a pod.
        """

        # Get when the pod started (reached the Kubelet) as a datetime
        start_time = getattr(getattr(pod, 'status', None), 'start_time', None)
        if start_time is None:
            # If the pod never made it to the kubelet to get a
            # start_time, say it was when the job was submitted.
            start_time = getattr(getattr(job, 'status', None), 'start_time', None)
        if start_time is None:
            # If this is still unset, say it was just now.
            start_time = utc_now()
        return start_time

    def getRunningBatchJobIDs(self) -> Dict[int, float]:
        # We need a dict from jobID (integer) to seconds it has been running
        secondsPerJob = dict()
        for job in self._ourJobObject():
            # Grab the pod for each job
            pod = self._getPodForJob(job)

            if pod is None:
                # Jobs whose pods are gone are not running
                continue

            if getattr(pod.status, 'phase', None) == 'Running':
                # The job's pod is running

                # Estimate the runtime
                runtime = (utc_now() - self._get_start_time(pod, job)).total_seconds()

                # Save it under the stringified job ID
                secondsPerJob[self._getIDForOurJob(job)] = runtime
        # Mix in the local jobs
        secondsPerJob.update(self.getRunningLocalJobIDs())
        return secondsPerJob

    def killBatchJobs(self, jobIDs: List[int]) -> None:

        # Kill all the ones that are local
        self.killLocalJobs(jobIDs)

        # Clears workflow's jobs listed in jobIDs.

        # First get the jobs we even issued non-locally
        issued_on_kubernetes = set(self._getIssuedNonLocalBatchJobIDs())
        deleted_jobs: List[str] = []

        for job_id in jobIDs:
            # For each job we are supposed to kill
            if job_id not in issued_on_kubernetes:
                # It never went to Kubernetes (or wasn't there when we just
                # looked), so we can't kill it on Kubernetes.

                # It is likely that the job is still waiting inside the jobs
                # queue and hasn't been submitted to Kubernetes. In this case,
                # it is not reliable to loop through the queue and try to
                # delete it. Instead, let's keep track of it and don't submit
                # when we encounter it.
                with self._mutex:
                    self._killed_queue_jobs.add(job_id)

                    # Make sure we don't keep track of it for getIssuedBatchJobIDs().
                    if job_id in self._queued_job_ids:
                        self._queued_job_ids.remove(job_id)

                continue
            # Work out what the job would be named
            job_name = self.job_prefix + str(job_id)

            # Delete the requested job in the foreground.
            # This doesn't block, but it does delete expeditiously.
            self._delete_job(job_name, propagation_policy='Foreground')

            deleted_jobs.append(job_name)
            logger.debug('Killed job by request: %s', job_name)

        for job_name in deleted_jobs:
            # Now we need to wait for all the jobs we killed to be gone.

            # Block until the delete takes.
            # The user code technically might stay running forever, but we push
            # the potential deadlock (if the user code needs exclusive access to
            # a resource) onto the user code, instead of always hanging
            # whenever we can't certify that a faulty node is no longer running
            # the user code. 
            self._waitForJobDeath(job_name)

    @classmethod
    def get_default_kubernetes_owner(cls) -> str:
        """
        Get the default Kubernetes-acceptable username string to tack onto jobs.
        """

        # Make a Kubernetes-acceptable version of our username: not too long,
        # and all lowercase letters, numbers, or - or .
        acceptable_chars = set(string.ascii_lowercase + string.digits + '-.')

        return ''.join([c for c in get_user_name().lower() if c in acceptable_chars])[:100]

    @runtime_checkable
    class KubernetesConfig(Protocol):
        """
        Type-enforcing protocol for Toil configs that have the extra Kubernetes
        batch system fields.

        TODO: Until MyPY lets protocols inherit form non-protocols, we will
        have to let the fact that this also has to be a Config just be manually
        enforced.
        """
        kubernetes_host_path: Optional[str]
        kubernetes_owner: str
        kubernetes_service_account: Optional[str]
        kubernetes_pod_timeout: float


    @classmethod
    def add_options(cls, parser: Union[ArgumentParser, _ArgumentGroup]) -> None:
        parser.add_argument("--kubernetesHostPath", dest="kubernetes_host_path", default=None,
                            help="Path on Kubernetes hosts to use as shared inter-pod temp directory.  "
                                 "(default: %(default)s)")
        parser.add_argument("--kubernetesOwner", dest="kubernetes_owner", default=cls.get_default_kubernetes_owner(),
                            help="Username to mark Kubernetes jobs with.  "
                                 "(default: %(default)s)")
        parser.add_argument("--kubernetesServiceAccount", dest="kubernetes_service_account", default=None,
                            help="Service account to run jobs as.  "
                                 "(default: %(default)s)")
        parser.add_argument("--kubernetesPodTimeout", dest="kubernetes_pod_timeout", default=120,
                            help="Seconds to wait for a scheduled Kubernetes pod to start running.  "
                                 "(default: %(default)s)")

    OptionType = TypeVar('OptionType')
    @classmethod
    def setOptions(cls, setOption: OptionSetter) -> None:
        setOption("kubernetes_host_path", default=None, env=['TOIL_KUBERNETES_HOST_PATH'])
        setOption("kubernetes_owner", default=cls.get_default_kubernetes_owner(), env=['TOIL_KUBERNETES_OWNER'])
        setOption("kubernetes_service_account", default=None, env=['TOIL_KUBERNETES_SERVICE_ACCOUNT'])
        setOption("kubernetes_pod_timeout", default=120, env=['TOIL_KUBERNETES_POD_TIMEOUT'])




# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import random
import shutil
import subprocess
import unittest
from contextlib import contextmanager
from uuid import uuid4

from toil import resolveEntryPoint
from toil.batchSystems.mesos.test import MesosTestSupport
from toil.common import Toil
from toil.job import Job
from toil.jobStores.abstractJobStore import (JobStoreExistsException,
                                             NoSuchJobStoreException)
from toil.exceptions import FailedJobsException
from toil.lib.bioio import root_logger
from toil.test import (ToilTest,
                       needs_aws_ec2,
                       needs_google,
                       needs_gridengine,
                       needs_mesos,
                       needs_parasol,
                       needs_torque,
                       slow)
from toil.test.batchSystems.parasolTestSupport import ParasolTestSupport
from toil.test.sort.sort import (copySubRangeOfFile,
                                 getMidPoint,
                                 main,
                                 makeFileToSort,
                                 merge,
                                 sort)

logger = logging.getLogger(__name__)

defaultLineLen = int(os.environ.get('TOIL_TEST_SORT_LINE_LEN', 10))
defaultLines = int(os.environ.get('TOIL_TEST_SORT_LINES', 10))
defaultN = int(os.environ.get('TOIL_TEST_SORT_N', defaultLineLen * defaultLines / 5))


@contextmanager
def runMain(options):
    """
    make sure the output file is deleted every time main is run
    """
    main(options)
    yield
    if os.path.exists(options.outputFile):
        os.remove(options.outputFile)


@slow
class SortTest(ToilTest, MesosTestSupport, ParasolTestSupport):
    """
    Tests Toil by sorting a file in parallel on various combinations of job stores and batch
    systems.
    """
    def setUp(self):
        super().setUp()
        self.tempDir = self._createTempDir(purpose='tempDir')
        self.outputFile = os.path.join(self.tempDir, 'sortedFile.txt')
        self.inputFile = os.path.join(self.tempDir, "fileToSort.txt")

    def tearDown(self):
        if os.path.exists(self.tempDir):
            shutil.rmtree(self.tempDir)
        ToilTest.tearDown(self)

    def _toilSort(self, jobStoreLocator, batchSystem,
                  lines=defaultLines, N=defaultN, testNo=1, lineLen=defaultLineLen,
                  retryCount=2, badWorker=0.5, downCheckpoints=False, caching=True):
        """
        Generate a file consisting of the given number of random lines, each line of the given
        length. Sort the file with Toil by splitting the file recursively until each part is less
        than the given number of bytes, sorting each part and merging them back together. Then
        verify the result.

        :param jobStoreLocator: The location of the job store.

        :param batchSystem: the name of the batch system

        :param lines: the number of random lines to generate

        :param N: the size in bytes of each split

        :param testNo: the number of repeats of this test

        :param lineLen: the length of each random line in the file
        """
        for test in range(testNo):
            try:
                # Specify options
                options = Job.Runner.getDefaultOptions(jobStoreLocator)
                options.logLevel = logging.getLevelName(root_logger.getEffectiveLevel())
                options.retryCount = retryCount
                options.batchSystem = batchSystem
                options.clean = "never"
                options.badWorker = badWorker
                options.badWorkerFailInterval = 0.05
                options.caching = caching
                # This is required because mesos_endpoint now defaults to the IP of the machine
                # that is starting the workflow while the mesos *tests* run locally.
                if batchSystem == 'mesos':
                    options.mesos_endpoint = 'localhost:5050'
                options.downCheckpoints = downCheckpoints
                options.N = N
                options.outputFile = self.outputFile
                options.fileToSort = self.inputFile
                options.overwriteOutput = True
                options.realTimeLogging = True

                # Make the file to sort
                makeFileToSort(options.fileToSort, lines=lines, lineLen=lineLen)

                # First make our own sorted version
                with open(options.fileToSort) as fileHandle:
                    l = fileHandle.readlines()
                    l.sort()

                # Check we get an exception if we try to restart a workflow that doesn't exist
                options.restart = True
                with self.assertRaises(NoSuchJobStoreException):
                    with runMain(options):
                        # Now check the file is properly sorted..
                        with open(options.outputFile) as fileHandle:
                            l2 = fileHandle.readlines()
                            self.assertEqual(l, l2)

                options.restart = False

                # Now actually run the workflow
                try:
                    with runMain(options):
                        pass
                    i = 0
                except FailedJobsException as e:
                    i = e.numberOfFailedJobs

                # Check we get an exception if we try to run without restart on an existing store
                with self.assertRaises(JobStoreExistsException):
                    with runMain(options):
                        pass

                options.restart = True

                # This loop tests the restart behavior
                totalTrys = 1
                while i != 0:
                    options.useExistingOptions = random.random() > 0.5
                    try:
                        with runMain(options):
                            pass
                        i = 0
                    except FailedJobsException as e:
                        i = e.numberOfFailedJobs
                        if totalTrys > 32:  # p(fail after this many restarts) = 0.5**32
                            self.fail('Exceeded a reasonable number of restarts')
                        totalTrys += 1
            finally:
                subprocess.check_call([resolveEntryPoint('toil'), 'clean', jobStoreLocator])
                # final test to make sure the jobStore was actually deleted
                self.assertRaises(NoSuchJobStoreException, Toil.resumeJobStore, jobStoreLocator)



    @needs_aws_ec2
    def testAwsSingle(self):
        self._toilSort(jobStoreLocator=self._awsJobStore(), batchSystem='single_machine')

    @needs_aws_ec2
    @needs_mesos
    def testAwsMesos(self):
        self._startMesos()
        try:
            self._toilSort(jobStoreLocator=self._awsJobStore(), batchSystem="mesos")
        finally:
            self._stopMesos()

    @needs_mesos
    def testFileMesos(self):
        self._startMesos()
        try:
            self._toilSort(jobStoreLocator=self._getTestJobStorePath(), batchSystem="mesos")
        finally:
            self._stopMesos()

    @needs_google
    def testGoogleSingle(self):
        self._toilSort(jobStoreLocator=self._googleJobStore(), batchSystem="single_machine")

    @needs_google
    @needs_mesos
    def testGoogleMesos(self):
        self._startMesos()
        try:
            self._toilSort(jobStoreLocator=self._googleJobStore(), batchSystem="mesos")
        finally:
            self._stopMesos()

    def testFileSingle(self):
        self._toilSort(jobStoreLocator=self._getTestJobStorePath(), batchSystem='single_machine')

    def testFileSingleNonCaching(self):
        self._toilSort(jobStoreLocator=self._getTestJobStorePath(), batchSystem='single_machine',
                       caching=False)

    def testFileSingleCheckpoints(self):
        self._toilSort(jobStoreLocator=self._getTestJobStorePath(), batchSystem='single_machine',
                       retryCount=2, downCheckpoints=True)

    def testFileSingle10000(self):
        self._toilSort(jobStoreLocator=self._getTestJobStorePath(), batchSystem='single_machine',
                       lines=10000, N=10000)

    @needs_gridengine
    @unittest.skip('GridEngine does not support shared caching')
    def testFileGridEngine(self):
        self._toilSort(jobStoreLocator=self._getTestJobStorePath(), batchSystem='gridengine')

    @needs_torque
    @unittest.skip('PBS/Torque does not support shared caching')
    def testFileTorqueEngine(self):
        self._toilSort(jobStoreLocator=self._getTestJobStorePath(), batchSystem='torque')

    @needs_parasol
    @unittest.skip("skipping until parasol support is less flaky (see github issue #449")
    def testFileParasol(self):
        self._startParasol()
        try:
            self._toilSort(jobStoreLocator=self._getTestJobStorePath(), batchSystem='parasol')
        finally:
            self._stopParasol()

    testNo = 5

    def testSort(self):
        for test in range(self.testNo):
            tempFile1 = os.path.join(self.tempDir, "fileToSort.txt")
            makeFileToSort(tempFile1)
            lines1 = self._loadFile(tempFile1)
            lines1.sort()
            sort(tempFile1)
            with open(tempFile1) as f:
                lines2 = f.readlines()
            self.assertEqual(lines1, lines2)

    def testMerge(self):
        for test in range(self.testNo):
            tempFile1 = os.path.join(self.tempDir, "fileToSort1.txt")
            tempFile2 = os.path.join(self.tempDir, "fileToSort2.txt")
            tempFile3 = os.path.join(self.tempDir, "mergedFile.txt")
            makeFileToSort(tempFile1)
            makeFileToSort(tempFile2)
            sort(tempFile1)
            sort(tempFile2)
            with open(tempFile3, 'w') as fileHandle:
                with open(tempFile1) as tempFileHandle1:
                    with open(tempFile2) as tempFileHandle2:
                        merge(tempFileHandle1, tempFileHandle2, fileHandle)
            lines1 = self._loadFile(tempFile1) + self._loadFile(tempFile2)
            lines1.sort()
            with open(tempFile3) as f:
                lines2 = f.readlines()
            self.assertEqual(lines1, lines2)

    def testCopySubRangeOfFile(self):
        for test in range(self.testNo):
            tempFile = os.path.join(self.tempDir, "fileToSort1.txt")
            outputFile = os.path.join(self.tempDir, "outputFileToSort1.txt")
            makeFileToSort(tempFile, lines=10, lineLen=defaultLineLen)
            fileSize = os.path.getsize(tempFile)
            assert fileSize > 0
            fileStart = random.choice(range(0, fileSize))
            fileEnd = random.choice(range(fileStart, fileSize))
            with open(outputFile, 'w') as f:
                f.write(copySubRangeOfFile(tempFile, fileStart, fileEnd))
            with open(outputFile) as f:
                l = f.read()
            with open(tempFile) as f:
                l2 = f.read()[fileStart:fileEnd]
            self.assertEqual(l, l2)

    def testGetMidPoint(self):
        for test in range(self.testNo):
            makeFileToSort(self.inputFile)
            with open(self.inputFile) as f:
                sorted_contents = f.read()
            fileSize = os.path.getsize(self.inputFile)
            midPoint = getMidPoint(self.inputFile, 0, fileSize)
            print(f"The mid point is {midPoint} of a file of {fileSize} bytes.")
            assert midPoint < fileSize
            assert sorted_contents[midPoint] == '\n'
            assert midPoint >= 0

    def _awsJobStore(self):
        return f'aws:{self.awsRegion()}:sort-test-{uuid4()}'

    def _googleJobStore(self):
        return f'google:{os.getenv("TOIL_GOOGLE_PROJECTID")}:sort-test-{uuid4()}'

    def _loadFile(self, path):
        with open(path) as f:
            return f.readlines()



# Copyright (C) 2022 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file contains WES utility functions borrowed from the Amazon Genomics
# CLI adapter codebase, at a6cc9904cd2d1899f7f64bf3c024a242c1a33749, to allow
# Toil to function as an AGC-compatible WES implementation without the use of a
# separate adapter. These functions help Toil use the WES semantics that AGC
# expects.

import json
import logging
import os
import sys
import zipfile
from os import path
from typing import IO, Any, Dict, List, Optional, Union, cast

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from urllib.parse import ParseResult, urlparse

from toil.bus import JobStatus
from toil.server.wes.abstract_backend import \
    MalformedRequestException as InvalidRequestError
from toil.server.wes.abstract_backend import TaskLog

logger = logging.getLogger(__name__)

# These functions are licensed under the same Apache 2.0 license as Toil is,
# but they come from a repo with a NOTICE file, so we must preserve this notice
# text. We can print it if Toil ever grows a function to print third-party
# license and notice text. Note that despite what this notice says, the Apache
# license claims that it doesn't constitute part of the licensing terms.
NOTICE = """
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
"""

# The license also requires a note that these functions have been modified by us.
# They needed to be modified to typecheck.

# The official spec we are working with here is: https://aws.github.io/amazon-genomics-cli/docs/concepts/workflows/#multi-file-workflows

class WorkflowPlan(TypedDict):
    """
    These functions pass around dicts of a certain type, with `data` and `files` keys.
    """
    data: "DataDict"
    files: "FilesDict"

class DataDict(TypedDict, total=False):
    """
    Under `data`, there can be:
    * `workflowUrl` (required if no `workflowSource`): URL to main workflow code.
    """
    workflowUrl: str

class FilesDict(TypedDict, total=False):
    """
    Under `files`, there can be:
    * `workflowSource` (required if no `workflowUrl`): Open binary-mode file for the main workflow code.
    * `workflowInputFiles`: List of open binary-mode file for input files. Expected to be JSONs.
    * `workflowOptions`: Open binary-mode file for a JSON of options sent along with the workflow.
    * `workflowDependencies`: Open binary-mode file for the zip the workflow came in, if any.
    """
    workflowSource: IO[bytes]
    workflowInputFiles: List[IO[bytes]]
    workflowOptions: IO[bytes]
    workflowDependencies: IO[bytes]

def parse_workflow_zip_file(file: str, workflow_type: str) -> WorkflowPlan:
    r"""
    Processes a workflow zip bundle

    :param file: String or Path-like path to a workflow.zip file
    :param workflow_type: String, extension of workflow to expect (e.g. "wdl")

    :rtype: dict of `data` and `files`

    If the zip only contains a single file, that file is set as `workflowSource`

    If the zip contains multiple files with a MANIFEST.json file, the MANIFEST is used to determine
    appropriate `data` and `file` arguments. (See: parse_workflow_manifest_file())

    If the zip contains multiple files without a MANIFEST.json file:
      * a `main` workflow file with an extension matching the workflow_type is expected and will be set as `workflowSource`
      * optionally, if `inputs*.json` files are found in the root level of the zip, they will be set as `workflowInputs(_\d)*` in the order they are found
      * optionally, if an `options.json` file is found in the root level of the zip, it will be set as `workflowOptions`

    If the zip contains multiple files, the original zip is set as `workflowDependencies`
    """
    data: DataDict = dict()
    files: FilesDict = dict()

    wd = path.dirname(file)
    with zipfile.ZipFile(file) as zip:
        zip.extractall(wd)

        contents = zip.namelist()
        if not contents:
            raise RuntimeError("empty workflow.zip")

        if len(contents) == 1:
            # single file workflow
            files["workflowSource"] = open(path.join(wd, contents[0]), "rb")

        else:
            # multifile workflow
            if "MANIFEST.json" in contents:
                props = parse_workflow_manifest_file(path.join(wd, "MANIFEST.json"))

                if props.get("data"):
                    data.update(props["data"])

                if props.get("files"):
                    files.update(props["files"])

            else:
                if not f"main.{workflow_type.lower()}" in contents:
                    raise RuntimeError(f"'main.{workflow_type}' file not found")

                files["workflowSource"] = open(
                    path.join(wd, f"main.{workflow_type.lower()}"), "rb"
                )

                input_files = [f for f in contents if f.startswith("inputs")]
                if input_files:
                    if not files.get("workflowInputFiles"):
                        files["workflowInputFiles"] = []

                    for input_file in input_files:
                        files[f"workflowInputFiles"] += [
                            open(path.join(wd, input_file), "rb")
                        ]

                if "options.json" in contents:
                    files["workflowOptions"] = open(path.join(wd, "options.json"), "rb")

            # add the original zip bundle as a workflow dependencies file
            files["workflowDependencies"] = open(file, "rb")

    return {"data": data, "files": files}


def parse_workflow_manifest_file(manifest_file: str) -> WorkflowPlan:
    r"""
    Reads a MANIFEST.json file for a workflow zip bundle

    :param manifest_file: String or Path-like path to a MANIFEST.json file

    :rtype: dict of `data` and `files`

    MANIFEST.json is expected to be formatted like:
    .. code-block:: json
       {
           "mainWorkflowURL": "relpath/to/workflow",
           "inputFileURLs": [
               "relpath/to/input-file-1",
               "relpath/to/input-file-2",
               ...
           ],
           "optionsFileURL" "relpath/to/option-file
       }

    The `mainWorkflowURL` property that provides a relative file path in the zip to a workflow file, which will be set as `workflowSource`

    The inputFileURLs property is optional and provides a list of relative file paths in the zip to input.json files. The list is assumed
    to be in the order the inputs should be applied - e.g. higher list index is higher priority. If present, it will be used to set
    `workflowInputs(_\d)` arguments.

    The optionsFileURL property is optional and  provides a relative file path in the zip to an options.json file. If present, it will be
    used to set `workflowOptions`.

    """
    data: DataDict = dict()
    files: FilesDict = dict()
    with open(manifest_file) as f:
        manifest = json.loads(f.read())

    u = urlparse(manifest["mainWorkflowURL"])
    if not u.scheme or u.scheme == "file":
        # expect "/path/to/file" or "file:///path/to/file"
        # root is relative to the zip root
        files["workflowSource"] = open(
            workflow_manifest_url_to_path(u, path.dirname(manifest_file)), "rb"
        )

    else:
        data["workflowUrl"] = manifest["mainWorkflowUrl"]

    if manifest.get("inputFileURLs"):
        if not files.get("workflowInputFiles"):
            files["workflowInputFiles"] = []

        for url in manifest["inputFileURLs"]:
            u = urlparse(url)
            if not u.scheme or u.scheme == "file":
                files[f"workflowInputFiles"] += [
                    open(
                        workflow_manifest_url_to_path(u, path.dirname(manifest_file)),
                        "rb",
                    )
                ]

            else:
                raise InvalidRequestError(
                    f"unsupported input file url scheme for: '{url}'"
                )

    if manifest.get("optionsFileURL"):
        u = urlparse(manifest["optionsFileURL"])
        if not u.scheme or u.scheme == "file":
            files["workflowOptions"] = open(
                workflow_manifest_url_to_path(u, path.dirname(manifest_file)), "rb"
            )
        else:
            raise InvalidRequestError(
                f"unsupported option file url scheme for: '{manifest['optionFileURL']}'"
            )

    return {"data": data, "files": files}


def workflow_manifest_url_to_path(url: ParseResult, parent_dir: Optional[str] = None) -> str:
    """
    Interpret a possibly-relative parsed URL, relative to the given parent directory.
    """
    relpath = url.path if not url.path.startswith("/") else url.path[1:]
    if parent_dir:
        return path.join(parent_dir, relpath)
    return relpath

# This one is all UCSC code
def task_filter(task: TaskLog, job_status: JobStatus) -> Optional[TaskLog]:
    """
    AGC requires task names to be annotated with an AWS Batch job ID that they
    were run under. If it encounters an un-annotated task name, it will crash.
    See <https://github.com/aws/amazon-genomics-cli/issues/494>.

    This encodes the AWSBatchJobID annotation, from the AmazonBatchBatchSystem,
    into the task name of the given task, and returns the modified task. If no
    such annotation is available, the task is censored and None is returned.
    """

    # Get the Batch ID for the task
    batch_id = job_status.external_batch_id

    if batch_id is None:
        return None

    modified_task = dict(task)
    # Tack the batch ID onto the end of the name with the required separator
    modified_task["name"] = "|".join([cast(str, modified_task.get("name", "")), batch_id])
    logger.info("Transformed task %s to %s", task, modified_task)
    return modified_task



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import logging
import os
import re
import struct
from shlex import quote

import requests

import docker
from docker.errors import (ContainerError,
                           ImageNotFound,
                           NotFound,
                           create_api_error_from_http_exception)
from docker.utils.socket import consume_socket_output, demux_adaptor

logger = logging.getLogger(__name__)

FORGO = 0
STOP = 1
RM = 2


def dockerCheckOutput(*args, **kwargs):
    raise RuntimeError("dockerCheckOutput() using subprocess.check_output() has been removed, "
                       "please switch to apiDockerCall().")


def dockerCall(*args, **kwargs):
    raise RuntimeError("dockerCall() using subprocess.check_output() has been removed, "
                       "please switch to apiDockerCall().")


def subprocessDockerCall(*args, **kwargs):
    raise RuntimeError("subprocessDockerCall() has been removed, "
                       "please switch to apiDockerCall().")


def apiDockerCall(job,
                  image,
                  parameters=None,
                  deferParam=None,
                  volumes=None,
                  working_dir=None,
                  containerName=None,
                  entrypoint=None,
                  detach=False,
                  log_config=None,
                  auto_remove=None,
                  remove=False,
                  user=None,
                  environment=None,
                  stdout=None,
                  stderr=False,
                  stream=False,
                  demux=False,
                  streamfile=None,
                  timeout=365 * 24 * 60 * 60,
                  **kwargs):
    """
    A toil wrapper for the python docker API.

    Docker API Docs: https://docker-py.readthedocs.io/en/stable/index.html
    Docker API Code: https://github.com/docker/docker-py

    This implements docker's python API within toil so that calls are run as
    jobs, with the intention that failed/orphaned docker jobs be handled
    appropriately.

    Example of using dockerCall in toil to index a FASTA file with SAMtools:
    def toil_job(job):
        working_dir = job.fileStore.getLocalTempDir()
        path = job.fileStore.readGlobalFile(ref_id,
                                          os.path.join(working_dir, 'ref.fasta')
        parameters = ['faidx', path]
        apiDockerCall(job,
                      image='quay.io/ucgc_cgl/samtools:latest',
                      working_dir=working_dir,
                      parameters=parameters)

    Note that when run with detach=False, or with detach=True and stdout=True
    or stderr=True, this is a blocking call. When run with detach=True and
    without output capture, the container is started and returned without
    waiting for it to finish.

    :param toil.Job.job job: The Job instance for the calling function.
    :param str image: Name of the Docker image to be used.
                     (e.g. 'quay.io/ucsc_cgl/samtools:latest')
    :param list[str] parameters: A list of string elements.  If there are
                                 multiple elements, these will be joined with
                                 spaces.  This handling of multiple elements
                                 provides backwards compatibility with previous
                                 versions which called docker using
                                 subprocess.check_call().
                                 **If list of lists: list[list[str]], then treat
                                 as successive commands chained with pipe.
    :param str working_dir: The working directory.
    :param int deferParam: Action to take on the container upon job completion.
           FORGO (0) leaves the container untouched and running.
           STOP (1) Sends SIGTERM, then SIGKILL if necessary to the container.
           RM (2) Immediately send SIGKILL to the container. This is the default
           behavior if deferParam is set to None.
    :param str name: The name/ID of the container.
    :param str entrypoint: Prepends commands sent to the container.  See:
                      https://docker-py.readthedocs.io/en/stable/containers.html
    :param bool detach: Run the container in detached mode. (equivalent to '-d')
    :param bool stdout: Return logs from STDOUT when detach=False (default: True).
                        Block and capture stdout to a file when detach=True
                        (default: False). Output capture defaults to output.log,
                        and can be specified with the "streamfile" kwarg.
    :param bool stderr: Return logs from STDERR when detach=False (default: False).
                        Block and capture stderr to a file when detach=True
                        (default: False). Output capture defaults to output.log,
                        and can be specified with the "streamfile" kwarg.
    :param bool stream: If True and detach=False, return a log generator instead
                        of a string. Ignored if detach=True. (default: False).
    :param bool demux: Similar to `demux` in container.exec_run(). If True and
                       detach=False, returns a tuple of (stdout, stderr). If
                       stream=True, returns a log generator with tuples of
                       (stdout, stderr). Ignored if detach=True. (default: False).
    :param str streamfile: Collect container output to this file if detach=True and
                        stderr and/or stdout are True. Defaults to "output.log".
    :param dict log_config: Specify the logs to return from the container.  See:
                      https://docker-py.readthedocs.io/en/stable/containers.html
    :param bool remove: Remove the container on exit or not.
    :param str user: The container will be run with the privileges of
                     the user specified.  Can be an actual name, such
                     as 'root' or 'lifeisaboutfishtacos', or it can be
                     the uid or gid of the user ('0' is root; '1000' is
                     an example of a less privileged uid or gid), or a
                     complement of the uid:gid (RECOMMENDED), such as
                     '0:0' (root user : root group) or '1000:1000'
                     (some other user : some other user group).
    :param environment: Allows one to set environment variables inside of the
                        container, such as:
    :param int timeout: Use the given timeout in seconds for interactions with
                        the Docker daemon. Note that the underlying docker module is
                        not always able to abort ongoing reads and writes in order
                        to respect the timeout. Defaults to 1 year (i.e. wait
                        essentially indefinitely).
    :param kwargs: Additional keyword arguments supplied to the docker API's
                   run command.  The list is 75 keywords total, for examples
                   and full documentation see:
                   https://docker-py.readthedocs.io/en/stable/containers.html

    :returns: Returns the standard output/standard error text, as requested, when
              detach=False. Returns the underlying
              docker.models.containers.Container object from the Docker API when
              detach=True.
    """

    # make certain that files have the correct permissions
    thisUser = os.getuid()
    thisGroup = os.getgid()
    if user is None:
        user = str(thisUser) + ":" + str(thisGroup)

    if containerName is None:
        containerName = getContainerName(job)

    if working_dir is None:
        working_dir = os.getcwd()

    if volumes is None:
        volumes = {working_dir: {'bind': '/data', 'mode': 'rw'}}

    for volume in volumes:
        if not os.path.exists(volume):
            os.makedirs(volume, exist_ok=True)

    if parameters is None:
        parameters = []

    # If 'parameters' is a list of lists, treat each list as a separate command
    # and chain with pipes.
    if len(parameters) > 0 and type(parameters[0]) is list:
        if entrypoint is None:
            entrypoint = ['/bin/bash', '-c']
        chain_params = \
            [' '.join(quote(arg) for arg in command) \
             for command in parameters]
        command = ' | '.join(chain_params)
        pipe_prefix = "set -eo pipefail && "
        command = [pipe_prefix + command]
        logger.debug("Calling docker with: " + repr(command))

    # If 'parameters' is a normal list, join all elements into a single string
    # element, quoting and escaping each element.
    # Example: ['echo','the Oread'] becomes: ["echo 'the Oread'"]
    # Note that this is still a list, and the docker API prefers this as best
    # practice:
    # http://docker-py.readthedocs.io/en/stable/containers.html
    elif len(parameters) > 0 and type(parameters) is list:
        command = ' '.join(quote(arg) for arg in parameters)
        logger.debug("Calling docker with: " + repr(command))

    # If the 'parameters' lists are empty, they are respecified as None, which
    # tells the API to simply create and run the container
    else:
        entrypoint = None
        command = None

    working_dir = os.path.abspath(working_dir)

    # Ensure the user has passed a valid value for deferParam
    assert deferParam in (None, FORGO, STOP, RM), \
        'Please provide a valid value for deferParam.'

    client = docker.from_env(version='auto', timeout=timeout)

    if deferParam is None:
        deferParam = RM

    if deferParam == STOP:
        job.defer(dockerStop, containerName)

    if deferParam == FORGO:
        # Leave the container untouched and running
        pass
    elif deferParam == RM:
        job.defer(dockerKill, containerName, remove=True)
    elif remove:
        job.defer(dockerKill, containerName, remove=True)

    if auto_remove is None:
        auto_remove = remove

    try:
        if detach is False:
            # When detach is False, this returns stdout normally:
            # >>> client.containers.run("ubuntu:latest", "echo hello world")
            # 'hello world\n'
            if stdout is None:
                stdout = True
            out = client.containers.run(image=image,
                                        command=command,
                                        working_dir=working_dir,
                                        entrypoint=entrypoint,
                                        name=containerName,
                                        detach=False,
                                        volumes=volumes,
                                        auto_remove=auto_remove,
                                        stdout=stdout,
                                        stderr=stderr,
                                        # to get the generator if demux=True
                                        stream=stream or demux,
                                        remove=remove,
                                        log_config=log_config,
                                        user=user,
                                        environment=environment,
                                        **kwargs)

            if demux is False:
                return out

            # If demux is True (i.e.: we want STDOUT and STDERR separated), we need to decode
            # the raw response from the docker API and preserve the stream type this time.
            response = out._response
            gen = (demux_adaptor(*frame) for frame in _multiplexed_response_stream_helper(response))

            if stream:
                return gen
            else:
                return consume_socket_output(frames=gen, demux=True)

        else:
            if (stdout or stderr) and log_config is None:
                logger.warning('stdout or stderr specified, but log_config is not set.  '
                               'Defaulting to "journald".')
                log_config = dict(type='journald')

            if stdout is None:
                stdout = False

            # When detach is True, this returns a container object:
            # >>> client.containers.run("bfirsh/reticulate-splines", detach=True)
            # <Container '45e6d2de7c54'>
            container = client.containers.run(image=image,
                                              command=command,
                                              working_dir=working_dir,
                                              entrypoint=entrypoint,
                                              name=containerName,
                                              detach=True,
                                              volumes=volumes,
                                              auto_remove=auto_remove,
                                              stdout=stdout,
                                              stderr=stderr,
                                              stream=stream,
                                              remove=remove,
                                              log_config=log_config,
                                              user=user,
                                              environment=environment,
                                              **kwargs)
            if stdout or stderr:
                if streamfile is None:
                    streamfile = 'output.log'
                with open(streamfile, 'wb') as f:
                    # stream=True makes this loop blocking; we will loop until
                    # the container stops and there is no more output.
                    for line in container.logs(stdout=stdout, stderr=stderr, stream=True):
                        f.write(line.encode() if isinstance(line, str) else line)

            # If we didn't capture output, the caller will need to .wait() on
            # the container to know when it is done. Even if we did capture
            # output, the caller needs the container to get at the exit code.
            return container

    except ContainerError:
        logger.error("Docker had non-zero exit.  Check your command: " + repr(command))
        raise
    except ImageNotFound:
        logger.error("Docker image not found.")
        raise
    except requests.exceptions.HTTPError as e:
        logger.error("The server returned an error.")
        raise create_api_error_from_http_exception(e)


def dockerKill(container_name: str,
               gentleKill: bool = False,
               remove: bool = False,
               timeout: int = 365 * 24 * 60 * 60) -> None:
    """
    Immediately kills a container.  Equivalent to "docker kill":
    https://docs.docker.com/engine/reference/commandline/kill/

    :param container_name: Name of the container being killed.
    :param gentleKill: If True, trigger a graceful shutdown.
    :param remove: If True, remove the container after it exits.
    :param int timeout: Use the given timeout in seconds for interactions with
                        the Docker daemon. Note that the underlying docker module is
                        not always able to abort ongoing reads and writes in order
                        to respect the timeout. Defaults to 1 year (i.e. wait
                        essentially indefinitely).
    """
    client = docker.from_env(version='auto', timeout=timeout)
    try:
        this_container = client.containers.get(container_name)
        while this_container.status == 'running':
            if gentleKill is False:
                client.containers.get(container_name).kill()
            else:
                client.containers.get(container_name).stop()
            this_container = client.containers.get(container_name)
        if remove:
            this_container.remove()
    except NotFound:
        logger.debug(f"Attempted to stop container ({container_name}), but container != exist.")
    except requests.exceptions.HTTPError as e:
        logger.debug(f"Attempted to stop container ({container_name}), but server gave an error:")
        raise create_api_error_from_http_exception(e)


def dockerStop(container_name: str, remove: bool = False) -> None:
    """
    Gracefully kills a container.  Equivalent to "docker stop":
    https://docs.docker.com/engine/reference/commandline/stop/

    :param container_name: Name of the container being stopped.
    :param remove: If True, remove the container after it exits.
    """
    dockerKill(container_name, gentleKill=True, remove=remove)


def containerIsRunning(container_name: str, timeout: int = 365 * 24 * 60 * 60):
    """
    Checks whether the container is running or not.

    :param container_name: Name of the container being checked.
    :param int timeout: Use the given timeout in seconds for interactions with
                        the Docker daemon. Note that the underlying docker module is
                        not always able to abort ongoing reads and writes in order
                        to respect the timeout. Defaults to 1 year (i.e. wait
                        essentially indefinitely).
    :returns: True if status is 'running', False if status is anything else,
    and None if the container does not exist.
    """
    client = docker.from_env(version='auto', timeout=timeout)
    try:
        this_container = client.containers.get(container_name)
        if this_container.status == 'running':
            return True
        else:
            # this_container.status == 'exited', 'restarting', or 'paused'
            return False
    except NotFound:
        return None
    except requests.exceptions.HTTPError as e:
        logger.debug("Server error attempting to call container: %s",
                     container_name)
        raise create_api_error_from_http_exception(e)


def getContainerName(job):
    """
    Create a random string including the job name, and return it. Name will
    match [a-zA-Z0-9][a-zA-Z0-9_.-]
    """
    parts = ['toil', str(job.description), base64.b64encode(os.urandom(9), b'-_').decode('utf-8')]
    name = re.sub('[^a-zA-Z0-9_.-]', '', '--'.join(parts))
    return name


def _multiplexed_response_stream_helper(response):
    """
    A generator of multiplexed data blocks coming from a response stream modified from:
    https://github.com/docker/docker-py/blob/4.3.1-release/docker/api/client.py#L370

    :param response: requests.Response
    :return: a generator with tuples of (stream_type, data)
    """
    while True:
        header = response.raw.read(8)
        if not header:
            break
        # header is 8 bytes with format: {STREAM_TYPE, 0, 0, 0, SIZE1, SIZE2, SIZE3, SIZE4}
        # protocol: https://docs.docker.com/engine/api/v1.24/#attach-to-a-container
        stream_type, length = struct.unpack('>BxxxL', header)
        if not length:
            continue
        data = response.raw.read(length)
        if not data:
            break
        yield stream_type, data



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Debug tool for running a toil job locally."""
import logging

from toil.common import Config, Toil, parser_with_common_options
from toil.statsAndLogging import set_logging_from_options
from toil.utils.toilDebugFile import printContentsOfJobStore
from toil.worker import workerScript

logger = logging.getLogger(__name__)


def main() -> None:
    parser = parser_with_common_options(jobstore_option=True)
    parser.add_argument("jobID", nargs=1,
                        help="The job store id of a job within the provided jobstore to run by itself.")
    parser.add_argument("--printJobInfo", nargs=1,
                        help="Return information about this job to the user including preceding jobs, "
                             "inputs, outputs, and runtime from the last known run.")

    options = parser.parse_args()
    set_logging_from_options(options)
    config = Config()
    config.setOptions(options)

    jobStore = Toil.resumeJobStore(config.jobStore)

    if options.printJobInfo:
        printContentsOfJobStore(jobStorePath=config.jobStore, nameOfJob=options.printJobInfo)

    # TODO: Option to print list of successor jobs
    # TODO: Option to run job within python debugger, allowing step through of arguments
    # idea would be to have option to import pdb and set breakpoint at the start of the user's code

    jobID = options.jobID[0]
    logger.debug(f"Running the following job locally: {jobID}")
    workerScript(jobStore, config, jobID, jobID, redirectOutputToLogFile=False)
    logger.debug(f"Finished running: {jobID}")



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import time

import toil.test.batchSystems.batchSystemTest as batchSystemTest
from toil.batchSystems.mesos.test import MesosTestSupport
from toil.job import Job, PromisedRequirement
from toil.lib.retry import retry_flaky_test
from toil.test import needs_mesos, slow

log = logging.getLogger(__name__)


class hidden:
    """
    Hide abstract base class from unittest's test case loader.

    http://stackoverflow.com/questions/1323455/python-unit-test-with-base-and-sub-class#answer-25695512
    """

    class AbstractPromisedRequirementsTest(batchSystemTest.hidden.AbstractBatchSystemJobTest):
        """An abstract base class for testing Toil workflows with promised requirements."""
        @slow
        def testConcurrencyDynamic(self):
            """
            Asserts that promised core resources are allocated properly using a dynamic Toil workflow
            """
            for coresPerJob in self.allocatedCores:
                log.debug('Testing %d cores per job with CPU count %d', coresPerJob, self.cpuCount)
                tempDir = self._createTempDir('testFiles')
                counterPath = self.getCounterPath(tempDir)

                root = Job.wrapJobFn(maxConcurrency, self.cpuCount, counterPath, coresPerJob,
                                     cores=1, memory='1M', disk='1M')
                values = Job.Runner.startToil(root, self.getOptions(tempDir))
                maxValue = max(values)
                self.assertEqual(maxValue, self.cpuCount // coresPerJob)

        @slow
        @retry_flaky_test(prepare=[batchSystemTest.hidden.AbstractBatchSystemJobTest.tearDown,
                                   batchSystemTest.hidden.AbstractBatchSystemJobTest.setUp])
        def testConcurrencyStatic(self):
            """
            Asserts that promised core resources are allocated properly using a static DAG
            """
            for coresPerJob in self.allocatedCores:
                log.debug('Testing %d cores per job with CPU count %d', coresPerJob, self.cpuCount)
                tempDir = self._createTempDir('testFiles')
                counterPath = self.getCounterPath(tempDir)

                root = Job()
                one = Job.wrapFn(getOne, cores=0.1, memory='32M', disk='1M')
                thirtyTwoMb = Job.wrapFn(getThirtyTwoMb, cores=0.1, memory='32M', disk='1M')
                root.addChild(one)
                root.addChild(thirtyTwoMb)
                for _ in range(self.cpuCount):
                    root.addFollowOn(Job.wrapFn(batchSystemTest.measureConcurrency, counterPath,
                                                cores=PromisedRequirement(lambda x: x * coresPerJob, one.rv()),
                                                memory=PromisedRequirement(thirtyTwoMb.rv()),
                                                disk='1M'))
                Job.Runner.startToil(root, self.getOptions(tempDir))
                _, maxValue = batchSystemTest.getCounters(counterPath)
                self.assertEqual(maxValue, self.cpuCount // coresPerJob)

        def getOptions(self, tempDir, caching=True):
            options = super().getOptions(tempDir)
            # defaultCores defaults to 1 - this is coincidentally the core requirement relied upon by this
            # test, so we change defaultCores to 2 to make the test more strict
            options.defaultCores = 2
            options.caching = caching
            return options

        def getCounterPath(self, tempDir):
            """
            Returns path to a counter file
            :param str tempDir: path to test directory
            :return: path to counter file
            """
            counterPath = os.path.join(tempDir, 'counter')
            batchSystemTest.resetCounters(counterPath)
            minValue, maxValue = batchSystemTest.getCounters(counterPath)
            assert (minValue, maxValue) == (0, 0)
            return counterPath

        def testPromisesWithJobStoreFileObjects(self, caching=True):
            """
            Check whether FileID objects are being pickled properly when used as return
            values of functions.  Then ensure that lambdas of promised FileID objects can be
            used to describe the requirements of a subsequent job.  This type of operation will be
            used commonly in Toil scripts.
            :return: None
            """
            file1 = 1024
            file2 = 512
            F1 = Job.wrapJobFn(_writer, file1)
            F2 = Job.wrapJobFn(_writer, file2)
            G = Job.wrapJobFn(_follower, file1 + file2,
                              disk=PromisedRequirement(lambda x, y: x.size + y.size,
                                                       F1.rv(), F2.rv()))
            F1.addChild(F2)
            F2.addChild(G)

            Job.Runner.startToil(F1, self.getOptions(self._createTempDir('testFiles'), caching=caching))

        def testPromisesWithNonCachingFileStore(self):
            self.testPromisesWithJobStoreFileObjects(caching=False)

        @slow
        def testPromiseRequirementRaceStatic(self):
            """
            Checks for a race condition when using promised requirements and child job functions.
            """
            A = Job.wrapJobFn(logDiskUsage, 'A', sleep=5, disk=PromisedRequirement(1024))
            B = Job.wrapJobFn(logDiskUsage, 'B', disk=PromisedRequirement(lambda x: x + 1024, A.rv()))
            A.addChild(B)
            Job.Runner.startToil(A, self.getOptions(self._createTempDir('testFiles')))


def _writer(job, fileSize):
    '''
    Write a local file and return the FileID obtained from running writeGlobalFile on
    it.

    :param job job: job
    :param int fileSize: Size of the file in bytes
    :returns: the result of writeGlobalFile on a locally created file
    :rtype: job.FileID
    '''
    with open(job.fileStore.getLocalTempFileName(), 'wb') as fH:
        fH.write(os.urandom(fileSize))
    return job.fileStore.writeGlobalFile(fH.name)


def _follower(job, expectedDisk):
    """
    This job follows the _writer jobs and ensures the expected disk is used.

    :param job job: job
    :param expectedDisk: Expect disk to be used by this job
    :return: None
    """
    assert job.disk == expectedDisk


def maxConcurrency(job, cpuCount, filename, coresPerJob):
    """
    Returns the max number of concurrent tasks when using a PromisedRequirement instance
    to allocate the number of cores per job.

    :param int cpuCount: number of available cpus
    :param str filename: path to counter file
    :param int coresPerJob: number of cores assigned to each job
    :return int max concurrency value:
    """
    one = job.addChildFn(getOne, cores=0.1, memory='32M', disk='1M')
    thirtyTwoMb = job.addChildFn(getThirtyTwoMb, cores=0.1, memory='32M', disk='1M')

    values = []
    for _ in range(cpuCount):
        value = job.addFollowOnFn(batchSystemTest.measureConcurrency, filename,
                                  cores=PromisedRequirement(lambda x: x * coresPerJob, one.rv()),
                                  memory=PromisedRequirement(thirtyTwoMb.rv()),
                                  disk='1M').rv()
        values.append(value)
    return values


def getOne():
    return 1


def getThirtyTwoMb():
    return '32M'


def logDiskUsage(job, funcName, sleep=0):
    """
    Logs the job's disk usage to master and sleeps for specified amount of time.

    :return: job function's disk usage
    """
    diskUsage = job.disk
    job.fileStore.logToMaster(f'{funcName}: {diskUsage}')
    time.sleep(sleep)
    return diskUsage


class SingleMachinePromisedRequirementsTest(hidden.AbstractPromisedRequirementsTest):
    """
    Tests against the SingleMachine batch system
    """

    def getBatchSystemName(self):
        return "single_machine"

    def tearDown(self):
        pass


@needs_mesos
class MesosPromisedRequirementsTest(hidden.AbstractPromisedRequirementsTest, MesosTestSupport):
    """
    Tests against the Mesos batch system
    """

    def getOptions(self, tempDir, caching=True):
        options = super().getOptions(tempDir, caching=caching)
        options.mesos_endpoint = 'localhost:5050'
        return options

    def getBatchSystemName(self):
        self._startMesos(self.cpuCount)
        return "mesos"

    def tearDown(self):
        self._stopMesos()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import imp
import os
from tempfile import NamedTemporaryFile

from setuptools import find_packages, setup


def get_requirements(extra=None):
    """
    Load the requirements for the given extra.

    Uses the appropriate requirements-extra.txt, or the main requirements.txt
    if no extra is specified.
    """
    filename = f"requirements-{extra}.txt" if extra else "requirements.txt"

    with open(filename) as fp:
        # Parse out as one per line, dropping comments
        return [l.split('#')[0].strip() for l in fp.readlines() if l.split('#')[0].strip()]


def run_setup():
    """
    Call setup().

    This function exists so the setup() invocation preceded more internal functionality.
    The `version` module is imported dynamically by import_version() below.
    """
    install_requires = get_requirements()

    extras_require = {}
    # htcondor is not supported by apple
    # this is tricky to conditionally support in 'all' due
    # to how wheels work, so it is not included in all and
    # must be explicitly installed as an extra
    all_reqs = []
    non_htcondor_extras = [
        "aws",
        "cwl",
        "encryption",
        "google",
        "kubernetes",
        "mesos",
        "wdl",
        "server"
    ]
    for extra in non_htcondor_extras:
        extras_require[extra] = get_requirements(extra)
        all_reqs += extras_require[extra]
    # We exclude htcondor from "all" because it can't be on Mac
    extras_require['htcondor:sys_platform!="darwin"'] = get_requirements("htcondor")
    extras_require["all"] = all_reqs

    setup(
        name='toil',
        version=version.distVersion,
        description='Pipeline management software for clusters.',
        author='Benedict Paten and the Toil community',
        author_email='toil-community@googlegroups.com',
        url="https://github.com/DataBiosphere/toil",
        classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Healthcare Industry',
          'License :: OSI Approved :: Apache Software License',
          'Natural Language :: English',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Scientific/Engineering :: Astronomy',
          'Topic :: Scientific/Engineering :: Atmospheric Science',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Medical Science Apps.',
          'Topic :: System :: Distributed Computing',
          'Topic :: Utilities'],
        license="Apache License v2.0",
        python_requires=">=3.7",
        install_requires=install_requires,
        extras_require=extras_require,
        package_dir={'': 'src'},
        packages=find_packages(where='src'),
        package_data={
            '': ['*.yml', '*.yaml', 'cloud-config', '*.cwl'],
        },
        # Unfortunately, the names of the entry points are hard-coded elsewhere in the code base so
        # you can't just change them here. Luckily, most of them are pretty unique strings, and thus
        # easy to search for.
        entry_points={
            'console_scripts': [
                'toil = toil.utils.toilMain:main',
                '_toil_worker = toil.worker:main',
                'cwltoil = toil.cwl.cwltoil:cwltoil_was_removed [cwl]',
                'toil-cwl-runner = toil.cwl.cwltoil:main [cwl]',
                'toil-wdl-runner = toil.wdl.wdltoil:main [wdl]',
                'toil-wdl-runner-old = toil.wdl.toilwdl:main [wdl]',
                'toil-wes-cwl-runner = toil.server.cli.wes_cwl_runner:main [server]',
                '_toil_mesos_executor = toil.batchSystems.mesos.executor:main [mesos]',
                '_toil_contained_executor = toil.batchSystems.contained_executor:executor']})


def import_version():
    """Return the module object for src/toil/version.py, generate from the template if required."""
    if not os.path.exists('src/toil/version.py'):
        for req in get_requirements("cwl"):
            # Determine cwltool version from requirements file
            if req.startswith("cwltool=="):
                cwltool_version = req[len("cwltool=="):]
                break
        # Use the template to generate src/toil/version.py
        import version_template
        with NamedTemporaryFile(mode='w', dir='src/toil', prefix='version.py.', delete=False) as f:
            f.write(version_template.expand_(others={
                # expose the dependency versions that we may need to access in Toil
                'cwltool_version': cwltool_version,
            }))
        os.rename(f.name, 'src/toil/version.py')

    # Unfortunately, we can't use a straight import here because that would also load the stuff
    # defined in "src/toil/__init__.py" which imports modules from external dependencies that may
    # yet to be installed when setup.py is invoked.
    #
    # This is also the reason we cannot switch from the "deprecated" imp library
    # and use:
    #     from importlib.machinery import SourceFileLoader
    #     return SourceFileLoader('toil.version', path='src/toil/version.py').load_module()
    #
    # Because SourceFileLoader will error and load "src/toil/__init__.py" .
    return imp.load_source('toil.version', 'src/toil/version.py')


version = import_version()
run_setup()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from toil.job import Job
from toil.test import ToilTest


class HelloWorldTest(ToilTest):
    def testHelloWorld(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.logLevel = "INFO"
        Job.Runner.startToil(HelloWorld(), options)

class HelloWorld(Job):
    def __init__(self):
        Job.__init__(self,  memory=100000, cores=2, disk="3G")

    def run(self, fileStore):
        fileID = self.addChildJobFn(childFn, cores=1, memory="1M", disk="3G").rv()
        self.addFollowOn(FollowOn(fileID))

def childFn(job):
    with job.fileStore.writeGlobalFileStream() as (fH, fileID):
        fH.write(b"Hello, World!")
        return fileID

class FollowOn(Job):
    def __init__(self,fileId):
        Job.__init__(self)
        self.fileId=fileId

    def run(self, fileStore):
        tempDir = fileStore.getLocalTempDir()
        tempFilePath = "/".join([tempDir,"LocalCopy"])
        with fileStore.readGlobalFileStream(self.fileId) as globalFile:
            with open(tempFilePath, "wb") as localFile:
                localFile.write(globalFile.read())



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from types import TracebackType
from typing import Any, ContextManager, List, Optional, Type

from toil.batchSystems.abstractBatchSystem import (BatchSystemSupport,
                                                   WorkerCleanupInfo)
from toil.batchSystems.local_support import BatchSystemLocalSupport
from toil.common import Config, Toil
from toil.lib.threading import LastProcessStandingArena

logger = logging.getLogger(__name__)

class BatchSystemCleanupSupport(BatchSystemLocalSupport):
    """
    Adds cleanup support when the last running job leaves a node, for batch
    systems that can't provide it using the backing scheduler.
    """

    @classmethod
    def supportsWorkerCleanup(cls) -> bool:
        return True

    def getWorkerContexts(self) -> List[ContextManager[Any]]:
        # Tell worker to register for and invoke cleanup

        # Create a context manager that has a copy of our cleanup info
        context = WorkerCleanupContext(self.workerCleanupInfo)

        # Send it along so the worker works inside of it
        contexts = super().getWorkerContexts()
        contexts.append(context)
        return contexts

    def __init__(self, config: Config, maxCores: float, maxMemory: int, maxDisk: int) -> None:
        super().__init__(config, maxCores, maxMemory, maxDisk)

class WorkerCleanupContext:
    """
    Context manager used by :class:`BatchSystemCleanupSupport` to implement
    cleanup on a node after the last worker is done working.

    Gets wrapped around the worker's work.
    """

    def __init__(self, workerCleanupInfo: WorkerCleanupInfo) -> None:
        """
        Wrap the given workerCleanupInfo in a context manager.

        :param workerCleanupInfo: Info to use to clean up the worker if we are
                                  the last to exit the context manager.
        """


        self.workerCleanupInfo = workerCleanupInfo
        # Don't set self.arena or MyPy will be upset that sometimes it doesn't have the right type.

    def __enter__(self) -> None:
        # Set up an arena so we know who is the last worker to leave
        self.arena = LastProcessStandingArena(Toil.get_toil_coordination_dir(self.workerCleanupInfo.work_dir, self.workerCleanupInfo.coordination_dir),
                                              self.workerCleanupInfo.workflow_id + '-cleanup')
        logger.debug('Entering cleanup arena')
        self.arena.enter()
        logger.debug('Cleanup arena entered')

    # This is exactly the signature MyPy demands.
    # Also, it demands we not say we can return a bool if we return False
    # always, because it can be smarter about reachability if it knows what
    # context managers never eat exceptions. So it decides any context manager
    # that is always falsey but claims to return a bool is an error.
    def __exit__(self, type: Optional[Type[BaseException]], value: Optional[BaseException], traceback: Optional[TracebackType]) -> None:
        logger.debug('Leaving cleanup arena')
        for _ in self.arena.leave():
            # We are the last concurrent worker to finish.
            # Do batch system cleanup.
            logger.debug('Cleaning up worker')
            BatchSystemSupport.workerCleanup(self.workerCleanupInfo)
            # Now the coordination_dir is allowed to no longer exist on the node.
        logger.debug('Cleanup arena left')


